# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Optional
import unified_planning as up
from unified_planning.exceptions import UPUsageError, UPValueError


class ROState:
    """This is an abstract class representing a classical Read Only state"""

    def get_value(self, value: "up.model.FNode") -> "up.model.FNode":
        """
        This method retrieves the value in the state.
        NOTE that the searched value must be set in the state.
        :params value: The value searched for in the state.
        :return: The set value.
        """
        raise NotImplementedError


class COWState(ROState):
    """This is an abstract class representing a classical Read/Copy-on-Write state"""

    def make_child(
        self, updated_values: Dict["up.model.FNode", "up.model.FNode"]
    ) -> "COWState":
        """
        Returns a different COWState in which every value in updated_values.keys() is evaluated as his mapping
        in new the updated_values dict and every other value is evaluated as in self.
        :param updated_values: The dictionary that contains the values that need to be updated in the new State.
        :return: The new State created.
        """
        raise NotImplementedError


class UPCOWState(COWState):
    """
    unified_planning implementation of the COWState interface.
    This class has a field "MAX_ANCESTORS" set to 20.

    The higher this numer is, the less memory the data structure will use.
    The lower this number is, the less time the data structure will need to retrieve a value.

    To set your own number just extend this class and re-define the MAX_ANCESTORS value. It must be > 0
    """

    MAX_ANCESTORS: int = 20
    """This class is the unified_planning implementation of the COWState interface."""

    def __init__(
        self,
        values: Dict["up.model.FNode", "up.model.FNode"],
        _father: Optional["UPCOWState"] = None,
    ):
        """
        Creates a new UPCOWState where the map values represents the get_value method. The parameter _father
        is for internal use only."""
        if type(self).MAX_ANCESTORS < 1:
            raise UPValueError(
                f"The max_ancestor field of a class extending UPCOWState must be > 0: in the class {type(self)} it is set to {type(self).MAX_ANCESTORS}"
            )
        self._father = _father
        self._values = values
        if _father is None:
            self._ancestors = 0
        else:
            self._ancestors = _father._ancestors + 1

    def __repr__(self) -> str:
        current_instance: Optional[UPCOWState] = self
        retval = []
        while current_instance is not None:
            retval.append(f"{str(current_instance._values)}")
            current_instance = current_instance._father
        return "\n".join(retval)

    def _has_local_value(self, value: "up.model.FNode") -> bool:
        """
        Method for internal use only.
        This method returns True if the parameter value is in this specific instance;
        NOTE that if self._has_local_value(x) returns False does not mean that self.get_value(x) will not retrieve any value.
        :param value: The value searched in this instance of the class.
        :return: True if this instance holds the value, False otherwise.
        """
        return value in self._values

    def get_value(self, value: "up.model.FNode") -> "up.model.FNode":
        """
        This method retrieves the value in the state.
        NOTE that the searched value must be set in the state.
        :params value: The value searched for in the state.
        :return: The set value.
        """
        right_instance: Optional[UPCOWState] = self
        while right_instance is not None:
            if right_instance._has_local_value(value):
                return right_instance._values[value]
            else:
                right_instance = right_instance._father
        raise UPUsageError(
            f"The state {self} does not have a value for the value {value}"
        )

    def make_child(
        self, updated_values: Dict["up.model.FNode", "up.model.FNode"]
    ) -> "UPCOWState":
        """
        Returns a different UPCOWState in which every value in updated_values.keys() is evaluated as his mapping
        in new the updated_values dict and every other value is evaluated as in self.
        :param updated_values: The dictionary that contains the values that need to be updated in the new State.
        :return: The new State created.
        """
        # If the number of ancestors is less that the given threshold it just creates a new state with self set as the father.
        if self._ancestors < type(self).MAX_ANCESTORS:
            return UPCOWState(updated_values, self)
        # Otherwise we retrieve every ancestor, and from the oldest to the newest we update the "complete_values" dict
        else:
            current_element: Optional[UPCOWState] = self
            complete_values = updated_values.copy()
            while current_element is not None:
                for k, v in current_element._values.items():
                    complete_values.setdefault(k, v)
                current_element = current_element._father
            return UPCOWState(complete_values)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning
from unified_planning.shortcuts import *
from unified_planning.model.problem_kind import (
    basic_classical_kind,
    classical_kind,
    full_numeric_kind,
    basic_temporal_kind,
    hierarchical_kind,
)
from unified_planning.test import (
    TestCase,
    skipIfNoPlanValidatorForProblemKind,
    skipIfNoOneshotPlannerForProblemKind,
    skipIfEngineNotAvailable,
)
from unified_planning.test.examples import get_example_problems
from unified_planning.engines import CompilationKind
from unified_planning.engines.compilers import Grounder


class TestGrounder(TestCase):
    def setUp(self):
        TestCase.setUp(self)
        self.problems = get_example_problems()

    def test_basic(self):
        problem = self.problems["basic"].problem

        gro = Grounder()

        res = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = res.problem
        res_2 = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem_2 = res_2.problem

        self.assertEqual(grounded_problem, grounded_problem_2)
        grounded_problem.name = problem.name
        self.assertEqual(grounded_problem, problem)

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot(self):
        problem = self.problems["robot"].problem

        gro = Grounder()
        res = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = res.problem
        self.assertEqual(len(grounded_problem.actions), 2)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = grounded_plan.replace_action_instances(res.map_back_action_instance)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected(self):
        problem = self.problems["robot_locations_connected"].problem

        gro = Grounder()
        res = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = res.problem
        self.assertEqual(len(grounded_problem.actions), 28)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = grounded_plan.replace_action_instances(res.map_back_action_instance)
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected_from_factory(self):
        problem = self.problems["robot_locations_connected"].problem

        with Compiler(name="up_grounder") as grounder:
            self.assertTrue(grounder.supports(problem.kind))
            res = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem = res.problem
            self.assertEqual(len(grounded_problem.actions), 28)
            for a in grounded_problem.actions:
                self.assertEqual(len(a.parameters), 0)

            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    res.map_back_action_instance
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(full_numeric_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(full_numeric_kind))
    def test_robot_locations_connected_from_factory_with_problem_kind(self):
        problem = self.problems["robot_locations_connected"].problem
        kind = problem.kind

        with Compiler(
            problem_kind=kind, compilation_kind=CompilationKind.GROUNDING
        ) as embedded_grounder:
            self.assertTrue(embedded_grounder.supports(kind))
            ground_result = embedded_grounder.compile(
                problem, CompilationKind.GROUNDING
            )
            grounded_problem, rewrite_plan_funct = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            self.assertEqual(len(grounded_problem.actions), 28)
            for a in grounded_problem.actions:
                self.assertEqual(len(a.parameters), 0)

            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(rewrite_plan_funct)
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(hierarchical_kind)
    @skipIfNoPlanValidatorForProblemKind(hierarchical_kind)
    def test_hierarchical_blocks_world(self):
        problem = self.problems["hierarchical_blocks_world"].problem

        gro = Grounder()
        ground_result = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = ground_result.problem
        self.assertEqual(len(grounded_problem.actions), 90)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = grounded_plan.replace_action_instances(
                ground_result.map_back_action_instance
            )
            for ai in plan.actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_matchcellar(self):
        problem = self.problems["matchcellar"].problem

        gro = Grounder()
        ground_result = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = ground_result.problem
        self.assertEqual(len(grounded_problem.actions), 6)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

        with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
            self.assertNotEqual(planner, None)
            grounded_plan = planner.solve(grounded_problem).plan
            plan = grounded_plan.replace_action_instances(
                ground_result.map_back_action_instance
            )
            for _, ai, _ in plan.timed_actions:
                a = ai.action
                self.assertEqual(a, problem.action(a.name))
            with PlanValidator(problem_kind=problem.kind, plan_kind=plan.kind) as pv:
                self.assertTrue(pv.validate(problem, plan))

    @skipIfNoOneshotPlannerForProblemKind(classical_kind.union(basic_temporal_kind))
    @skipIfNoPlanValidatorForProblemKind(classical_kind.union(basic_temporal_kind))
    def test_matchcellar_grounder_from_factory(self):
        problem = self.problems["matchcellar"].problem

        gro = Grounder()
        ground_result = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem_test = ground_result.problem
        with Compiler(name="up_grounder") as grounder:
            self.assertTrue(grounder.supports(problem.kind))
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem_try, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            self.assertEqual(grounded_problem_test, grounded_problem_try)
            with OneshotPlanner(problem_kind=grounded_problem_try.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem_try).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for _, ai, _ in plan.timed_actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    def test_timed_connected_locations(self):
        problem = self.problems["timed_connected_locations"].problem

        gro = Grounder()
        ground_result = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = ground_result.problem
        self.assertEqual(len(grounded_problem.actions), 20)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

    def test_ad_hoc_1(self):
        problem = Problem("ad_hoc")
        Location = UserType("Location")
        visited = Fluent("at", BoolType(), position=Location)
        l1 = Object("l1", Location)
        visit = InstantaneousAction("visit", l_to=Location)
        l_to = visit.parameter("l_to")
        visit.add_effect(visited(l_to), True)
        visit_l1 = InstantaneousAction("visit_l1")
        visit_l1.add_effect(visited(l1), True)
        problem.add_fluent(visited)
        problem.set_initial_value(visited(l1), True)
        problem.add_object(l1)
        problem.add_action(visit)
        problem.add_action(visit_l1)
        gro = Grounder()
        ground_result = gro.compile(problem, CompilationKind.GROUNDING)
        grounded_problem = ground_result.problem
        self.assertEqual(len(grounded_problem.actions), 2)
        for a in grounded_problem.actions:
            self.assertEqual(len(a.parameters), 0)

    @skipIfEngineNotAvailable("pyperplan")
    def test_pyperplan_grounder(self):
        problem = self.problems["robot_no_negative_preconditions"].problem
        for action in problem.actions:
            self.assertTrue(len(action.parameters) > 0)
        with Compiler(name="pyperplan") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))

    @skipIfEngineNotAvailable("pyperplan")
    def test_pyperplan_grounder_mockup_problem(self):
        problem = Problem("mockup")
        Location = UserType("Location")
        at = Fluent("at", BoolType(), position=Location)
        at_l2 = Fluent("at_l2")
        l1 = Object("l1", Location)
        l2 = Object("l2", Location)
        move_to = InstantaneousAction("move_to", l_to=Location)
        l_to = move_to.parameter("l_to")
        move_to.add_effect(at(l_to), True)
        move_to_l2 = InstantaneousAction("move_to_l2")
        move_to_l2.add_effect(at_l2, True)
        problem.add_fluent(at, default_initial_value=False)
        problem.add_fluent(at_l2, default_initial_value=False)
        problem.add_object(l1)
        problem.add_object(l2)
        problem.add_action(move_to)
        problem.add_action(move_to_l2)
        problem.add_goal(at(l1))
        problem.add_goal(at(l2))
        problem.add_goal(at_l2)

        with Compiler(name="pyperplan") as grounder:
            ground_result = grounder.compile(problem, CompilationKind.GROUNDING)
            grounded_problem, rewrite_back_plan_function = (
                ground_result.problem,
                ground_result.map_back_action_instance,
            )
            for grounded_action in grounded_problem.actions:
                self.assertEqual(len(grounded_action.parameters), 0)
            with OneshotPlanner(problem_kind=grounded_problem.kind) as planner:
                self.assertNotEqual(planner, None)
                grounded_plan = planner.solve(grounded_problem).plan
                plan = grounded_plan.replace_action_instances(
                    rewrite_back_plan_function
                )
                for ai in plan.actions:
                    a = ai.action
                    self.assertEqual(a, problem.action(a.name))
                with PlanValidator(
                    problem_kind=problem.kind, plan_kind=plan.kind
                ) as pv:
                    self.assertTrue(pv.validate(problem, plan))



#!/usr/bin/env python3

from setuptools import setup, find_packages  # type: ignore
import unified_planning


long_description = """============================================================
 Unified planning: A library that unifies planning frameworks
 ============================================================
    Insert long description here
"""

setup(
    name="unified_planning",
    version=unified_planning.__version__,
    description="Unified Planning Framework",
    author="AIPlan4EU Project",
    author_email="aiplan4eu@fbk.eu",
    url="https://www.aiplan4eu-project.eu",
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7",  # supported Python ranges
    install_requires=["pyparsing", "networkx"],
    extras_require={
        "dev": ["tarski[arithmetic]", "pytest", "pytest-cov", "mypy"],
        "grpc": ["grpcio", "grpcio-tools", "grpc-stubs"],
        "tarski": ["tarski[arithmetic]"],
        "pyperplan": ["up-pyperplan==0.2.0"],
        "tamer": ["up-tamer==0.2.0.14.dev1"],
        "enhsp": ["up-enhsp==0.0.6"],
        "fast-downward": ["up-fast-downward==0.0.4"],
        "engines": [
            "tarski[arithmetic]",
            "up-pyperplan==0.2.0",
            "up-tamer==0.2.0.14.dev1",
            "up-enhsp==0.0.6",
            "up-fast-downward==0.0.4",
        ],
    },
    license="APACHE",
    keywords="planning logic STRIPS RDDL",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from unified_planning.model.types import _UserType
from unified_planning.exceptions import UPProblemDefinitionError, UPValueError
from typing import List, Dict, Optional, cast


class UserTypesSetMixin:
    """
    This class is a mixin that contains a set of user types with some related methods.

    NOTE: when this mixin is used in combination with other mixins that share some
    of the attributes (e.g. has_name_method), it is required to pass the very same
    arguments to the mixins constructors.
    """

    def __init__(self, has_name_method):
        self._has_name_method = has_name_method
        self._user_types: List["up.model.types.Type"] = []
        # The field _user_types_hierarchy stores the information about the types and the list of their sons.
        self._user_types_hierarchy: Dict[
            Optional["up.model.types.Type"], List["up.model.types.Type"]
        ] = {}

    def _add_user_type(self, type: "up.model.types.Type"):
        """This method adds a Type, together with all it's ancestors, to the user_types_hierarchy"""
        assert type.is_user_type()
        if type not in self._user_types:
            t = cast(_UserType, type)
            if self._has_name_method(t.name):
                raise UPProblemDefinitionError(
                    f"The type name {t.name} is already used in the problem"
                )
            if t.father is not None:
                self._add_user_type(t.father)
            self._user_types.append(type)

    @property
    def user_types(self) -> List["up.model.types.Type"]:
        """Returns the user types."""
        return self._user_types

    def user_type(self, name: str) -> "up.model.types.Type":
        """Returns the user type with the given name."""
        for ut in self.user_types:
            assert ut.is_user_type()
            if cast(_UserType, ut).name == name:
                return ut
        raise UPValueError(f"UserType {name} is not defined!")

    def has_type(self, name: str) -> bool:
        """Returns True iff the type 'name' is defined."""
        for ut in self.user_types:
            assert ut.is_user_type()
            if cast(_UserType, ut).name == name:
                return True
        return False

    @property
    def user_types_hierarchy(
        self,
    ) -> Dict[Optional["up.model.types.Type"], List["up.model.types.Type"]]:
        """Returns a Dict where every key represents an Optional Type and the value
        associated to the key is the List of the direct sons of the Optional Type."""
        res: Dict[Optional["up.model.types.Type"], List["up.model.types.Type"]] = {}
        for t in self._user_types:
            if t not in res:
                res[t] = []
            f = cast(_UserType, t).father
            if f not in res:
                res[f] = [t]
            else:
                res[f].append(t)
        return res



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the conditional effects remover class."""


import unified_planning as up
import unified_planning.engines as engines
from unified_planning.engines.mixins.compiler import CompilationKind, CompilerMixin
from unified_planning.engines.results import CompilerResult
from unified_planning.exceptions import (
    UPProblemDefinitionError,
    UPConflictingEffectsException,
)
from unified_planning.model import Problem, ProblemKind
from unified_planning.engines.compilers.utils import (
    get_fresh_name,
    check_and_simplify_preconditions,
    check_and_simplify_conditions,
    replace_action,
)
from unified_planning.utils import powerset
from typing import Iterable, List, Dict, Tuple
from functools import partial


class ConditionalEffectsRemover(engines.engine.Engine, CompilerMixin):
    """Conditional effect remover class: this class offers the capability
    to transform a conditional problem into an unconditional one.

    This is done by substituting every conditional action with different
    actions representing every possible branch of the original action.

    Also the conditional timed_effects are removed maintaining the same
    semanthics. When this is not possible, an exception is raised."""

    @property
    def name(self):
        return "cerm"

    @staticmethod
    def supported_kind() -> ProblemKind:
        supported_kind = ProblemKind()
        supported_kind.set_problem_class("ACTION_BASED")
        supported_kind.set_typing("FLAT_TYPING")
        supported_kind.set_typing("HIERARCHICAL_TYPING")
        supported_kind.set_numbers("CONTINUOUS_NUMBERS")
        supported_kind.set_numbers("DISCRETE_NUMBERS")
        supported_kind.set_fluents_type("NUMERIC_FLUENTS")
        supported_kind.set_fluents_type("OBJECT_FLUENTS")
        supported_kind.set_conditions_kind("NEGATIVE_CONDITIONS")
        supported_kind.set_conditions_kind("DISJUNCTIVE_CONDITIONS")
        supported_kind.set_conditions_kind("EQUALITY")
        supported_kind.set_conditions_kind("EXISTENTIAL_CONDITIONS")
        supported_kind.set_conditions_kind("UNIVERSAL_CONDITIONS")
        supported_kind.set_effects_kind("CONDITIONAL_EFFECTS")
        supported_kind.set_effects_kind("INCREASE_EFFECTS")
        supported_kind.set_effects_kind("DECREASE_EFFECTS")
        supported_kind.set_time("CONTINUOUS_TIME")
        supported_kind.set_time("DISCRETE_TIME")
        supported_kind.set_time("INTERMEDIATE_CONDITIONS_AND_EFFECTS")
        supported_kind.set_time("TIMED_EFFECT")
        supported_kind.set_time("TIMED_GOALS")
        supported_kind.set_time("DURATION_INEQUALITIES")
        supported_kind.set_simulated_entities("SIMULATED_EFFECTS")
        return supported_kind

    @staticmethod
    def supports(problem_kind):
        return problem_kind <= ConditionalEffectsRemover.supported_kind()

    @staticmethod
    def supports_compilation(compilation_kind: CompilationKind) -> bool:
        return compilation_kind == CompilationKind.CONDITIONAL_EFFECTS_REMOVING

    def _compile(
        self,
        problem: "up.model.AbstractProblem",
        compilation_kind: "up.engines.CompilationKind",
    ) -> CompilerResult:
        assert isinstance(problem, Problem)

        env = problem.env
        simplifier = env.simplifier

        new_to_old = {}

        new_problem = problem.clone()
        new_problem.name = f"{self.name}_{problem.name}"
        new_problem.clear_timed_effects()
        for t, el in problem.timed_effects.items():
            for e in el:
                if e.is_conditional():
                    f, v = e.fluent.fluent(), e.value
                    if not f.type.is_bool_type():
                        raise UPProblemDefinitionError(
                            f"The condition of effect: {e}\ncould not be removed without changing the problem."
                        )
                    else:
                        em = env.expression_manager
                        c = e.condition
                        nv = simplifier.simplify(
                            em.Or(em.And(c, v), em.And(em.Not(c), f))
                        )
                        new_problem.add_timed_effect(t, e.fluent, nv)
                else:
                    new_problem._add_effect_instance(t, e.clone())
        new_problem.clear_actions()
        for ua in problem.unconditional_actions:
            new_uncond_action = ua.clone()
            new_problem.add_action(new_uncond_action)
            new_to_old[new_uncond_action] = ua
        for action in problem.conditional_actions:
            if isinstance(action, up.model.InstantaneousAction):
                cond_effects = action.conditional_effects
                for p in powerset(range(len(cond_effects))):
                    new_action = action.clone()
                    new_action.name = get_fresh_name(new_problem, action.name)
                    new_action.clear_effects()
                    for e in action.unconditional_effects:
                        new_action._add_effect_instance(e.clone())
                    for i, e in enumerate(cond_effects):
                        if i in p:
                            # positive precondition
                            new_action.add_precondition(e.condition)
                            ne = up.model.Effect(
                                e.fluent, e.value, env.expression_manager.TRUE(), e.kind
                            )
                            # We try to add the new effect, but it might be in conflict with exising effects,
                            # so the action is not added to the problem
                            try:
                                new_action._add_effect_instance(ne)
                            except UPConflictingEffectsException:
                                continue
                        else:
                            # negative precondition
                            new_action.add_precondition(
                                env.expression_manager.Not(e.condition)
                            )
                    # new action is created, then is checked if it has any impact and if it can be simplified
                    if len(new_action.effects) > 0:
                        (
                            action_is_feasible,
                            simplified_preconditions,
                        ) = check_and_simplify_preconditions(
                            new_problem, new_action, simplifier
                        )
                        if action_is_feasible:
                            new_action._set_preconditions(simplified_preconditions)
                            new_to_old[new_action] = action
                            new_problem.add_action(new_action)
            elif isinstance(action, up.model.DurativeAction):
                timing_cond_effects: Dict[
                    "up.model.Timing", List["up.model.Effect"]
                ] = action.conditional_effects
                cond_effects_timing: List[
                    Tuple["up.model.Effect", "up.model.Timing"]
                ] = [(e, t) for t, el in timing_cond_effects.items() for e in el]
                for p in powerset(range(len(cond_effects_timing))):
                    new_action = action.clone()
                    new_action.name = get_fresh_name(new_problem, action.name)
                    new_action.clear_effects()
                    for t, el in action.unconditional_effects.items():
                        for e in el:
                            new_action._add_effect_instance(t, e.clone())
                    for i, (e, t) in enumerate(cond_effects_timing):
                        if i in p:
                            # positive precondition
                            new_action.add_condition(t, e.condition)
                            ne = up.model.Effect(
                                e.fluent, e.value, env.expression_manager.TRUE(), e.kind
                            )
                            # We try to add the new effect, but it might be in conflict with exising effects,
                            # so the action is not added to the problem
                            try:
                                new_action._add_effect_instance(t, ne)
                            except UPConflictingEffectsException:
                                continue
                        else:
                            # negative precondition
                            new_action.add_condition(
                                t, env.expression_manager.Not(e.condition)
                            )
                    # new action is created, then is checked if it has any impact and if it can be simplified
                    if len(new_action.effects) > 0:
                        (
                            action_is_feasible,
                            simplified_conditions,
                        ) = check_and_simplify_conditions(
                            new_problem, new_action, simplifier
                        )
                        if action_is_feasible:
                            new_action.clear_conditions()
                            for interval, c in simplified_conditions:
                                new_action.add_condition(interval, c)
                            new_to_old[new_action] = action
                            new_problem.add_action(new_action)
            else:
                raise NotImplementedError

        return CompilerResult(
            new_problem, partial(replace_action, map=new_to_old), self.name
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module defines the different remover classes."""


import unified_planning as up
from unified_planning.exceptions import UPConflictingEffectsException
from unified_planning.model import (
    FNode,
    TimeInterval,
    Action,
    InstantaneousAction,
    DurativeAction,
    Problem,
    Effect,
    Expression,
    SimulatedEffect,
    Parameter,
)
from unified_planning.plans import ActionInstance
from typing import Dict, Iterable, List, Optional, Tuple


def check_and_simplify_conditions(
    problem: Problem, action: DurativeAction, simplifier
) -> Tuple[bool, List[Tuple[TimeInterval, FNode]]]:
    """
    Simplifies conditions and if it is False (a contraddiction)
    returns False, otherwise returns True.
    If the simplification is True (a tautology) removes all conditions at the given timing.
    If the simplification is still an AND rewrites back every "arg" of the AND
    in the conditions
    If the simplification is not an AND sets the simplification as the only
    condition at the given timing.
    Then, the new conditions are returned as a List[Tuple[Timing, FNode]] and the user can
    decide how to use the new conditions.
    """
    # new action conditions
    nac: List[Tuple[TimeInterval, FNode]] = []
    # t = timing, lc = list condition
    for i, lc in action.conditions.items():
        # conditions (as an And FNode)
        c = problem.env.expression_manager.And(lc)
        # conditions simplified
        cs = simplifier.simplify(c)
        if cs.is_bool_constant():
            if not cs.bool_constant_value():
                return (
                    False,
                    [],
                )
        else:
            if cs.is_and():
                for new_cond in cs.args:
                    nac.append((i, new_cond))
            else:
                nac.append((i, cs))
    return (True, nac)


def check_and_simplify_preconditions(
    problem: Problem, action: InstantaneousAction, simplifier
) -> Tuple[bool, List[FNode]]:
    """
    Simplifies preconditions and if it is False (a contraddiction)
    returns False, otherwise returns True.
    If the simplification is True (a tautology) removes all preconditions.
    If the simplification is still an AND rewrites back every "arg" of the AND
    in the preconditions
    If the simplification is not an AND sets the simplification as the only
    precondition.
    Then, the new preconditions are returned as a List[FNode] and the user can
    decide how to use the new preconditions.
    """
    # action preconditions
    ap = action.preconditions
    if len(ap) == 0:
        return (True, [])
    # preconditions (as an And FNode)
    p = problem.env.expression_manager.And(ap)
    # preconditions simplified
    ps = simplifier.simplify(p)
    # new action preconditions
    nap: List[FNode] = []
    if ps.is_bool_constant():
        if not ps.bool_constant_value():
            return (False, [])
    else:
        if ps.is_and():
            nap.extend(ps.args)
        else:
            nap.append(ps)
    action._set_preconditions(nap)
    return (True, nap)


def create_effect_with_given_subs(
    problem: Problem,
    old_effect: Effect,
    simplifier,
    substituter,
    subs: Dict[Expression, Expression],
) -> Optional[Effect]:
    new_fluent = substituter.substitute(old_effect.fluent, subs)
    new_value = substituter.substitute(old_effect.value, subs)
    new_condition = simplifier.simplify(
        substituter.substitute(old_effect.condition, subs)
    )
    if new_condition == problem.env.expression_manager.FALSE():
        return None
    else:
        return Effect(new_fluent, new_value, new_condition, old_effect.kind)


def create_action_with_given_subs(
    problem: Problem,
    old_action: Action,
    simplifier,
    substituter,
    subs: Dict[Expression, Expression],
) -> Optional[Action]:
    naming_list: List[str] = []
    for param, value in subs.items():
        assert isinstance(param, Parameter)
        assert isinstance(value, FNode)
        naming_list.append(str(value))
    if isinstance(old_action, InstantaneousAction):
        new_action = InstantaneousAction(
            get_fresh_name(problem, old_action.name, naming_list)
        )
        for p in old_action.preconditions:
            new_action.add_precondition(substituter.substitute(p, subs))
        for e in old_action.effects:
            new_effect = create_effect_with_given_subs(
                problem, e, simplifier, substituter, subs
            )
            if new_effect is not None:
                # We try to add the new effect, but a compiler might generate conflicting effects,
                # so the action is just considered invalid
                try:
                    new_action._add_effect_instance(new_effect)
                except UPConflictingEffectsException:
                    return None
        se = old_action.simulated_effect
        if se is not None:
            new_fluents = []
            for f in se.fluents:
                new_fluents.append(substituter.substitute(f, subs))

            def fun(_problem, _state, _):
                return se.function(_problem, _state, subs)

            # We try to add the new simulated effect, but a compiler might generate conflicting effects,
            # so the action is just considered invalid
            try:
                new_action.set_simulated_effect(SimulatedEffect(new_fluents, fun))
            except UPConflictingEffectsException:
                return None
        is_feasible, new_preconditions = check_and_simplify_preconditions(
            problem, new_action, simplifier
        )
        if not is_feasible:
            return None
        new_action._set_preconditions(new_preconditions)
        return new_action
    elif isinstance(old_action, DurativeAction):
        new_durative_action = DurativeAction(
            get_fresh_name(problem, old_action.name, naming_list)
        )
        new_durative_action.set_duration_constraint(old_action.duration)
        for i, cl in old_action.conditions.items():
            for c in cl:
                new_durative_action.add_condition(i, substituter.substitute(c, subs))
        for t, el in old_action.effects.items():
            for e in el:
                new_effect = create_effect_with_given_subs(
                    problem, e, simplifier, substituter, subs
                )
                if new_effect is not None:
                    # We try to add the new simulated effect, but a compiler might generate conflicting effects,
                    # so the action is just considered invalid
                    try:
                        new_durative_action._add_effect_instance(t, new_effect)
                    except UPConflictingEffectsException:
                        return None
        for t, se in old_action.simulated_effects.items():
            new_fluents = []
            for f in se.fluents:
                new_fluents.append(substituter.substitute(f, subs))

            def fun(_problem, _state, _):
                return se.function(_problem, _state, subs)

            # We try to add the new simulated effect, but a compiler might generate conflicting effects,
            # so the action is just considered invalid
            try:
                new_durative_action.set_simulated_effect(
                    t, SimulatedEffect(new_fluents, fun)
                )
            except UPConflictingEffectsException:
                return None
        is_feasible, new_conditions = check_and_simplify_conditions(
            problem, new_durative_action, simplifier
        )
        if not is_feasible:
            return None
        new_durative_action.clear_conditions()
        for interval, c in new_conditions:
            new_durative_action.add_condition(interval, c)
        return new_durative_action
    else:
        raise NotImplementedError


def get_fresh_name(
    problem: Problem, original_name: str, parameters_names: Iterable[str] = []
) -> str:
    """This method returns a fresh name for the problem, given a name and an iterable of names in input."""
    name_list = [original_name]
    name_list.extend(parameters_names)
    new_name = "_".join(name_list)
    base_name = new_name
    count = 0
    while problem.has_name(new_name):
        new_name = f"{base_name}_{str(count)}"
        count += 1
    return new_name


def lift_action_instance(
    action_instance: ActionInstance,
    map: Dict["up.model.Action", Tuple["up.model.Action", List["up.model.FNode"]]],
) -> ActionInstance:
    """ "map" is a map from every action in the "grounded_problem" to the tuple
    (original_action, parameters).

    Where the grounded action is obtained by grounding
    the "original_action" with the specific "parameters"."""
    lifted_action, parameters = map[action_instance.action]
    return ActionInstance(lifted_action, tuple(parameters))


def replace_action(
    action_instance: ActionInstance, map: Dict["up.model.Action", "up.model.Action"]
) -> ActionInstance:
    return ActionInstance(
        map[action_instance.action], action_instance.actual_parameters
    )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# type: ignore[attr-defined]
from functools import partial
from typing import Tuple, Union, Optional
import fractions
from typing import OrderedDict

import unified_planning.grpc.generated.unified_planning_pb2 as proto
from unified_planning.exceptions import UPException
from unified_planning import Environment
from unified_planning import model
from unified_planning.model import metrics
import unified_planning.plans
from unified_planning.grpc.converter import Converter, handles
from unified_planning.model import (
    DurativeAction,
    Effect,
    InstantaneousAction,
    Parameter,
    Problem,
    Variable,
)
from unified_planning.model.effect import EffectKind
from unified_planning.model.operators import OperatorKind


def convert_type_str(s: str, problem: Problem) -> model.types.Type:
    if s == "up:bool":
        return problem.env.type_manager.BoolType()
    elif s == "up:integer":
        return problem.env.type_manager.IntType()
    elif "up:integer[" in s:
        lb = int(s.split("[")[1].split(",")[0])
        ub = int(s.split(",")[1].split("]")[0])
        return problem.env.type_manager.IntType(lb, ub)
    elif s == "up:real":
        return problem.env.type_manager.RealType()
    elif "up:real[" in s:
        return problem.env.type_manager.RealType(
            lower_bound=fractions.Fraction(s.split("[")[1].split(",")[0]),
            upper_bound=fractions.Fraction(s.split(",")[1].split("]")[0]),
        )
    else:
        assert not s.startswith("up:"), f"Unhandled builtin type: {s}"
        return problem.user_type(s)


# The operators are based on SExpressions supported in PDDL.
def op_to_node_type(op: str) -> OperatorKind:
    if op == "up:plus":
        return OperatorKind.PLUS
    elif op == "up:minus":
        return OperatorKind.MINUS
    elif op == "up:times":
        return OperatorKind.TIMES
    elif op == "up:div":
        return OperatorKind.DIV
    elif op == "up:equals":
        return OperatorKind.EQUALS
    elif op == "up:le":
        return OperatorKind.LE
    elif op == "up:lt":
        return OperatorKind.LT
    elif op == "up:and":
        return OperatorKind.AND
    elif op == "up:or":
        return OperatorKind.OR
    elif op == "up:not":
        return OperatorKind.NOT
    elif op == "up:exists":
        return OperatorKind.EXISTS
    elif op == "up:forall":
        return OperatorKind.FORALL
    elif op == "up:implies":
        return OperatorKind.IMPLIES
    elif op == "up:iff":
        return OperatorKind.IFF

    raise ValueError(f"Unknown operator `{op}`")


class ProtobufReader(Converter):
    @handles(proto.Parameter)
    def _convert_parameter(
        self, msg: proto.Parameter, problem: Problem
    ) -> model.Parameter:
        return model.Parameter(
            msg.name, convert_type_str(msg.type, problem), problem.env
        )

    @handles(proto.Fluent)
    def _convert_fluent(self, msg: proto.Fluent, problem: Problem) -> model.Fluent:
        value_type: model.types.Type = convert_type_str(msg.value_type, problem)
        sig: list = []
        for p in msg.parameters:
            sig.append(self.convert(p, problem))
        fluent = model.Fluent(msg.name, value_type, sig, problem.env)
        return fluent

    @handles(proto.ObjectDeclaration)
    def _convert_object(
        self, msg: proto.ObjectDeclaration, problem: Problem
    ) -> model.Object:
        return model.Object(msg.name, convert_type_str(msg.type, problem))

    @handles(proto.Expression)
    def _convert_expression(
        self, msg: proto.Expression, problem: Problem
    ) -> model.Expression:
        if msg.kind == proto.ExpressionKind.Value("CONSTANT"):
            assert msg.atom is not None
            return self.convert(msg.atom, problem)

        elif msg.kind == proto.ExpressionKind.Value("PARAMETER"):
            return problem.env.expression_manager.ParameterExp(
                param=Parameter(
                    msg.atom.symbol, convert_type_str(msg.type, problem), problem.env
                ),
            )
        elif msg.kind == proto.ExpressionKind.Value("VARIABLE"):
            return problem.env.expression_manager.VariableExp(
                var=Variable(
                    msg.atom.symbol, convert_type_str(msg.type, problem), problem.env
                ),
            )
        elif msg.kind == proto.ExpressionKind.Value("STATE_VARIABLE"):
            args = []
            payload = None

            fluent = msg.list.pop(0)
            if fluent.kind == proto.ExpressionKind.Value("FLUENT_SYMBOL"):
                payload = self.convert(fluent.atom, problem)

            args.extend([self.convert(m, problem) for m in msg.list])
            if payload is not None:
                return problem.env.expression_manager.FluentExp(payload, tuple(args))
            else:
                raise UPException(f"Unable to form fluent expression {msg}")
        elif (
            msg.kind == proto.ExpressionKind.Value("FUNCTION_APPLICATION")
            and msg.type != "up:time"
        ):
            node_type = None
            args = []
            payload = None

            symbol = msg.list.pop(0)
            if symbol.kind == proto.ExpressionKind.Value("FUNCTION_SYMBOL"):
                node_type = op_to_node_type(symbol.atom.symbol)

            if node_type in [OperatorKind.EXISTS, OperatorKind.FORALL]:
                variables = msg.list[:-1]
                quantified_expression = msg.list[-1]
                args.append(self.convert(quantified_expression, problem))
                payload = tuple(
                    [self.convert(var, problem).variable() for var in variables]
                )
            else:
                args.extend([self.convert(m, problem) for m in msg.list])

            assert node_type is not None

            return problem.env.expression_manager.create_node(
                node_type=node_type,
                args=tuple(args),
                payload=payload,
            )
        elif (
            msg.kind == proto.ExpressionKind.Value("FUNCTION_APPLICATION")
            and msg.type == "up:time"
        ):
            fn = msg.list[0].atom.symbol
            if fn == "up:start":
                kd = model.TimepointKind.START
            elif fn == "up:end":
                kd = model.TimepointKind.END
            elif fn == "up:global_start":
                kd = model.TimepointKind.GLOBAL_START
            elif fn == "up:global_end":
                kd = model.TimepointKind.GLOBAL_END
            else:
                raise ValueError(f"Invalid temporal qualifier {fn}")
            container = None
            if len(msg.list) > 1:
                container = msg.list[1].atom.symbol
            tp = model.timing.Timepoint(kd, container)
            return problem.env.expression_manager.TimingExp(model.Timing(0, tp))

        raise ValueError(f"Unknown expression kind `{msg.kind}`")

    @handles(proto.Atom)
    def _convert_atom(
        self, msg: proto.Atom, problem: Problem
    ) -> Union[model.FNode, model.Fluent, model.Object]:
        field = msg.WhichOneof("content")

        value = getattr(msg, field)
        if field == "int":
            return problem.env.expression_manager.Int(value)
        elif field == "real":
            return problem.env.expression_manager.Real(
                fractions.Fraction(value.numerator, value.denominator)
            )
        elif field == "boolean":
            return problem.env.expression_manager.Bool(value)
        else:
            # If atom symbols, return the equivalent UP alternative
            # Note that parameters are directly handled at expression level
            if problem.has_object(value):
                return problem.env.expression_manager.ObjectExp(
                    obj=problem.object(value)
                )
            else:
                return problem.fluent(value)

    @handles(proto.TypeDeclaration)
    def _convert_type_declaration(
        self, msg: proto.TypeDeclaration, problem: Problem
    ) -> model.Type:
        if msg.type_name == "up:bool":
            return problem.env.type_manager.BoolType()
        elif msg.type_name.startswith("up:integer["):
            tmp = msg.type_name.split("[")[1].split("]")[0].split(", ")
            return problem.env.type_manager.IntType(
                lower_bound=int(tmp[0]) if tmp[0] != "-inf" else None,
                upper_bound=int(tmp[1]) if tmp[1] != "inf" else None,
            )
        elif msg.type_name.startswith("up:real["):
            tmp = msg.type_name.split("[")[1].split("]")[0].split(", ")
            lower_bound = fractions.Fraction(tmp[0]) if tmp[0] != "-inf" else None
            upper_bound = fractions.Fraction(tmp[1]) if tmp[1] != "inf" else None
            return problem.env.type_manager.RealType(
                lower_bound=lower_bound, upper_bound=upper_bound
            )
        else:
            father = (
                problem.user_type(msg.parent_type) if msg.parent_type != "" else None
            )
            return problem.env.type_manager.UserType(name=msg.type_name, father=father)

    @handles(proto.Problem)
    def _convert_problem(
        self, msg: proto.Problem, env: Optional[Environment] = None
    ) -> Problem:
        problem_name = str(msg.problem_name) if str(msg.problem_name) != "" else None
        if msg.HasField("hierarchy"):
            problem = model.htn.HierarchicalProblem(name=problem_name, env=env)
        else:
            problem = Problem(name=problem_name, env=env)

        for t in msg.types:
            problem._add_user_type(self.convert(t, problem))
        for obj in msg.objects:
            problem.add_object(self.convert(obj, problem))
        for f in msg.fluents:
            problem.add_fluent(
                self.convert(f, problem),
                default_initial_value=self.convert(f.default_value, problem)
                if f.HasField("default_value")
                else None,
            )
        for f in msg.actions:
            problem.add_action(self.convert(f, problem))
        for eff in msg.timed_effects:
            ot = self.convert(eff.occurrence_time, problem)
            effect = self.convert(eff.effect, problem)
            problem.add_timed_effect(
                timing=ot,
                fluent=effect.fluent,
                value=effect.value,
                condition=effect.condition,
            )

        for assign in msg.initial_state:
            problem.set_initial_value(
                fluent=self.convert(assign.fluent, problem),
                value=self.convert(assign.value, problem),
            )

        for g in msg.goals:
            goal = self.convert(g.goal, problem)
            if str(g.timing) == "":
                problem.add_goal(goal)
            else:
                timing = self.convert(g.timing)
                problem.add_timed_goal(interval=timing, goal=goal)

        for metric in msg.metrics:
            problem.add_quality_metric(self.convert(metric, problem))

        if msg.HasField("hierarchy"):
            for task in msg.hierarchy.abstract_tasks:
                problem.add_task(self.convert(task, problem))
            for method in msg.hierarchy.methods:
                problem.add_method(self.convert(method, problem))
            problem._initial_task_network = self.convert(
                msg.hierarchy.initial_task_network, problem
            )

        return problem

    @handles(proto.AbstractTaskDeclaration)
    def _convert_abstract_task(
        self, msg: proto.AbstractTaskDeclaration, problem: Problem
    ):
        return model.htn.Task(
            msg.name, [self.convert(p, problem) for p in msg.parameters], problem.env
        )

    @handles(proto.Task)
    def _convert_task(
        self, msg: proto.Task, problem: model.htn.HierarchicalProblem
    ) -> model.htn.Subtask:
        if problem.has_task(msg.task_name):
            task = problem.get_task(msg.task_name)
        elif problem.has_action(msg.task_name):
            task = problem.action(msg.task_name)
        else:
            raise ValueError(f"Unknown task name: {msg.task_name}")
        parameters = [self.convert(p, problem) for p in msg.parameters]
        return model.htn.Subtask(task, *parameters, ident=msg.id, _env=problem.env)

    @handles(proto.Method)
    def _convert_method(
        self, msg: proto.Method, problem: model.htn.HierarchicalProblem
    ) -> model.htn.Method:
        method = model.htn.Method(
            msg.name,
            [self.convert(p, problem) for p in msg.parameters],
            problem.env,
        )
        achieved_task_params = []
        for p in msg.achieved_task.parameters:
            achieved_task_params.append(method.parameter(p.atom.symbol))
        method.set_task(
            problem.get_task(msg.achieved_task.task_name), *achieved_task_params
        )
        for st in msg.subtasks:
            method.add_subtask(self.convert(st, problem))
        for c in msg.constraints:
            method.add_constraint(self.convert(c, problem))
        for c in msg.conditions:
            assert not c.HasField("span"), "Timed conditions are currently unsupported."
            method.add_precondition(self.convert(c.cond, problem))
        return method

    @handles(proto.TaskNetwork)
    def _convert_task_network(
        self, msg: proto.TaskNetwork, problem: model.htn.HierarchicalProblem
    ) -> model.htn.TaskNetwork:
        tn = model.htn.TaskNetwork(problem.env)
        for v in msg.variables:
            tn.add_variable(v.name, convert_type_str(v.type, problem))
        for st in msg.subtasks:
            tn.add_subtask(self.convert(st, problem))
        for c in msg.constraints:
            tn.add_constraint(self.convert(c, problem))

        return tn

    @handles(proto.Metric)
    def _convert_metric(
        self, msg: proto.Metric, problem: Problem
    ) -> Union[
        metrics.MinimizeActionCosts,
        metrics.MinimizeSequentialPlanLength,
        metrics.MinimizeMakespan,
        metrics.MinimizeExpressionOnFinalState,
        metrics.MaximizeExpressionOnFinalState,
        metrics.Oversubscription,
    ]:
        if msg.kind == proto.Metric.MINIMIZE_ACTION_COSTS:
            costs = {}
            for a, cost in msg.action_costs.items():
                costs[problem.action(a)] = self.convert(cost, problem)
            return metrics.MinimizeActionCosts(
                costs=costs,
                default=self.convert(msg.default_action_cost, problem)
                if msg.HasField("default_action_cost")
                else None,
            )

        elif msg.kind == proto.Metric.MINIMIZE_SEQUENTIAL_PLAN_LENGTH:
            return metrics.MinimizeSequentialPlanLength()

        elif msg.kind == proto.Metric.MINIMIZE_MAKESPAN:
            return metrics.MinimizeMakespan()

        elif msg.kind == proto.Metric.MINIMIZE_EXPRESSION_ON_FINAL_STATE:
            return metrics.MinimizeExpressionOnFinalState(
                expression=self.convert(msg.expression, problem)
            )

        elif msg.kind == proto.Metric.MAXIMIZE_EXPRESSION_ON_FINAL_STATE:
            return metrics.MaximizeExpressionOnFinalState(
                expression=self.convert(msg.expression, problem)
            )
        elif msg.kind == proto.Metric.OVERSUBSCRIPTION:
            goals = {}
            for g in msg.goals:
                goals[self.convert(g.goal, problem)] = self.convert(g.cost)
            return metrics.Oversubscription(goals)
        else:
            raise UPException(f"Unknown metric kind `{msg.kind}`")

    @handles(proto.Action)
    def _convert_action(self, msg: proto.Action, problem: Problem) -> model.Action:
        action: model.Action

        parameters = OrderedDict()
        for param in msg.parameters:
            parameters[param.name] = convert_type_str(param.type, problem)

        if msg.HasField("duration"):
            action = DurativeAction(msg.name, parameters)
            action.set_duration_constraint(self.convert(msg.duration, problem))
        else:
            action = InstantaneousAction(msg.name, parameters)

        conditions = []
        for condition in msg.conditions:
            cond = self.convert(condition.cond, problem)
            span = self.convert(condition.span) if condition.HasField("span") else None
            conditions.append((cond, span))

        effects = []
        for effect in msg.effects:
            eff = self.convert(effect.effect, problem)
            time = (
                self.convert(effect.occurrence_time)
                if effect.HasField("occurrence_time")
                else None
            )
            effects.append((eff, time))

        if isinstance(action, DurativeAction):
            for c, span in conditions:
                action.add_condition(span, c)
            for e, ot in effects:
                if e.kind == EffectKind.ASSIGN:
                    action.add_effect(ot, e.fluent, e.value, e.condition)
                elif e.kind == EffectKind.DECREASE:
                    action.add_decrease_effect(ot, e.fluent, e.value, e.condition)
                elif e.kind == EffectKind.INCREASE:
                    action.add_increase_effect(ot, e.fluent, e.value, e.condition)
        elif isinstance(action, InstantaneousAction):
            for c, _ in conditions:
                action.add_precondition(c)
            for e, _ in effects:
                if e.kind == EffectKind.ASSIGN:
                    action.add_effect(e.fluent, e.value, e.condition)
                elif e.kind == EffectKind.DECREASE:
                    action.add_decrease_effect(e.fluent, e.value, e.condition)
                elif e.kind == EffectKind.INCREASE:
                    action.add_increase_effect(e.fluent, e.value, e.condition)

        return action

    @handles(proto.EffectExpression)
    def _convert_effect(
        self, msg: proto.EffectExpression, problem: Problem
    ) -> model.Effect:
        # EffectKind
        if msg.kind == proto.EffectExpression.EffectKind.Value("INCREASE"):
            kind = EffectKind.INCREASE
        elif msg.kind == proto.EffectExpression.EffectKind.Value("DECREASE"):
            kind = EffectKind.DECREASE
        else:
            kind = EffectKind.ASSIGN

        return Effect(
            fluent=self.convert(msg.fluent, problem),
            value=self.convert(msg.value, problem),
            condition=self.convert(msg.condition, problem),
            kind=kind,
        )

    @handles(proto.Duration)
    def _convert_duration(
        self, msg: proto.Duration, problem: Problem
    ) -> model.timing.DurationInterval:
        return model.timing.DurationInterval(
            lower=self.convert(msg.controllable_in_bounds.lower, problem),
            upper=self.convert(msg.controllable_in_bounds.upper, problem),
            is_left_open=bool(msg.controllable_in_bounds.is_left_open),
            is_right_open=bool(msg.controllable_in_bounds.is_right_open),
        )

    @handles(proto.TimeInterval)
    def _convert_timed_interval(self, msg: proto.TimeInterval) -> model.TimeInterval:
        return model.TimeInterval(
            lower=self.convert(msg.lower),
            upper=self.convert(msg.upper),
            is_left_open=msg.is_left_open,
            is_right_open=msg.is_right_open,
        )

    @handles(proto.Timing)
    def _convert_timing(self, msg: proto.Timing) -> model.timing.Timing:
        return model.Timing(
            delay=self.convert(msg.delay)
            if msg.HasField("delay")
            else fractions.Fraction(0),
            timepoint=self.convert(msg.timepoint),
        )

    @handles(proto.Real)
    def _convert_real(self, msg: proto.Real) -> fractions.Fraction:
        return fractions.Fraction(msg.numerator, msg.denominator)

    @handles(proto.Timepoint)
    def _convert_timepoint(self, msg: proto.Timepoint) -> model.timing.Timepoint:
        if msg.kind == proto.Timepoint.TimepointKind.Value("GLOBAL_START"):
            kind = model.timing.TimepointKind.GLOBAL_START
        elif msg.kind == proto.Timepoint.TimepointKind.Value("GLOBAL_END"):
            kind = model.timing.TimepointKind.GLOBAL_END
        elif msg.kind == proto.Timepoint.TimepointKind.Value("START"):
            kind = model.timing.TimepointKind.START
        elif msg.kind == proto.Timepoint.TimepointKind.Value("END"):
            kind = model.timing.TimepointKind.END
        else:
            raise UPException("Unknown timepoint kind: {}".format(msg.kind))
        container = msg.container_id if msg.container_id != "" else None
        return model.timing.Timepoint(kind, container)

    @handles(proto.Plan)
    def _convert_plan(
        self, msg: proto.Plan, problem: Problem
    ) -> unified_planning.plans.Plan:
        actions = [self.convert(a, problem) for a in msg.actions]
        if all(isinstance(a, tuple) for a in actions):
            # If all actions are tuples, we can assume that they are
            # (absolute start time, action, duration)
            return unified_planning.plans.TimeTriggeredPlan(actions)
        else:
            # Otherwise, we assume they are instantenous actions
            return unified_planning.plans.SequentialPlan(actions=actions)

    @handles(proto.ActionInstance)
    def _convert_action_instance(
        self, msg: proto.ActionInstance, problem: Problem
    ) -> Union[
        Tuple[
            model.timing.Timing,
            unified_planning.plans.ActionInstance,
            model.timing.Duration,
        ],
        unified_planning.plans.ActionInstance,
    ]:
        # action instance parameters are atoms but in UP they are FNodes
        # converting to up.model.FNode
        parameters = tuple([self.convert(param, problem) for param in msg.parameters])

        action_instance = unified_planning.plans.ActionInstance(
            problem.action(msg.action_name),
            parameters,
        )

        start_time = (
            self.convert(msg.start_time) if msg.HasField("start_time") else None
        )
        end_time = self.convert(msg.end_time) if msg.HasField("end_time") else None
        if start_time is not None:
            return (
                start_time,  # Absolute Start Time
                action_instance,
                end_time - start_time if end_time else None,  # Duration
            )
        else:
            return action_instance

    @handles(proto.PlanGenerationResult)
    def _convert_plan_generation_result(
        self, result: proto.PlanGenerationResult, problem: Problem
    ) -> unified_planning.engines.PlanGenerationResult:
        if result.status == proto.PlanGenerationResult.Status.Value(
            "SOLVED_SATISFICING"
        ):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.SOLVED_SATISFICING
            )
        elif result.status == proto.PlanGenerationResult.Status.Value(
            "SOLVED_OPTIMALLY"
        ):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.SOLVED_OPTIMALLY
            )
        elif result.status == proto.PlanGenerationResult.Status.Value(
            "UNSOLVABLE_PROVEN"
        ):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.UNSOLVABLE_PROVEN
            )
        elif result.status == proto.PlanGenerationResult.Status.Value(
            "UNSOLVABLE_INCOMPLETELY"
        ):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY
            )
        elif result.status == proto.PlanGenerationResult.Status.Value("TIMEOUT"):
            status = unified_planning.engines.results.PlanGenerationResultStatus.TIMEOUT
        elif result.status == proto.PlanGenerationResult.Status.Value("MEMOUT"):
            status = unified_planning.engines.results.PlanGenerationResultStatus.MEMOUT
        elif result.status == proto.PlanGenerationResult.Status.Value("INTERNAL_ERROR"):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.INTERNAL_ERROR
            )
        elif result.status == proto.PlanGenerationResult.Status.Value(
            "UNSUPPORTED_PROBLEM"
        ):
            status = (
                unified_planning.engines.results.PlanGenerationResultStatus.UNSUPPORTED_PROBLEM
            )
        else:
            raise UPException(f"Unknown Planner Status: {result.status}")

        log_messages = None
        metrics = None

        if bool(result.metrics):
            metrics = dict(result.metrics)

        if len(result.log_messages) > 0:
            log_messages = [self.convert(log) for log in result.log_messages]

        return unified_planning.engines.PlanGenerationResult(
            status=status,
            plan=self.convert(result.plan, problem),
            engine_name=result.engine.name,
            metrics=metrics,
            log_messages=log_messages,
        )

    @handles(proto.LogMessage)
    def _convert_log_message(
        self, log: proto.LogMessage
    ) -> unified_planning.engines.LogMessage:
        if log.level == proto.LogMessage.LogLevel.Value("INFO"):
            return unified_planning.engines.LogMessage(
                level=unified_planning.engines.LogLevel.INFO,
                message=log.message,
            )
        elif log.level == proto.LogMessage.LogLevel.Value("WARNING"):
            return unified_planning.engines.LogMessage(
                level=unified_planning.engines.LogLevel.WARNING,
                message=log.message,
            )
        elif log.level == proto.LogMessage.LogLevel.Value("ERROR"):
            return unified_planning.engines.LogMessage(
                level=unified_planning.engines.LogLevel.ERROR,
                message=log.message,
            )
        elif log.level == proto.LogMessage.LogLevel.Value("DEBUG"):
            return unified_planning.engines.LogMessage(
                level=unified_planning.engines.LogLevel.DEBUG,
                message=log.message,
            )
        else:
            raise UPException(f"Unexpected Log Level: {log.level}")

    @handles(proto.CompilerResult)
    def _convert_compiler_result(
        self,
        result: proto.CompilerResult,
        lifted_problem: unified_planning.model.Problem,
    ) -> unified_planning.engines.CompilerResult:
        problem = self.convert(result.problem, lifted_problem.env)
        map: Dict[
            unified_planning.model.Action,
            Tuple[unified_planning.model.Action, List[unified_planning.model.FNode]],
        ] = {}
        for grounded_action in problem.actions:
            original_action_instance = self.convert(
                result.map_back_plan[grounded_action.name], lifted_problem
            )
            map[grounded_action] = (
                original_action_instance.action,
                original_action_instance.actual_parameters,
            )
        return unified_planning.engines.CompilerResult(
            problem=problem,
            map_back_action_instance=partial(
                unified_planning.engines.compilers.utils.lift_action_instance, map=map
            ),
            engine_name=result.engine.name,
            log_messages=[self.convert(log) for log in result.log_messages],
        )

    @handles(proto.ValidationResult)
    def _convert_validation_result(
        self, result: proto.ValidationResult
    ) -> unified_planning.engines.ValidationResult:
        if result.status == proto.ValidationResult.ValidationResultStatus.Value(
            "VALID"
        ):
            r_status = unified_planning.engines.ValidationResultStatus.VALID
        elif result.status == proto.ValidationResult.ValidationResultStatus.Value(
            "INVALID"
        ):
            r_status = unified_planning.engines.ValidationResultStatus.INVALID
        else:
            raise UPException(f"Unexpected ValidationResult status: {result.status}")
        return unified_planning.engines.ValidationResult(
            status=r_status,
            engine_name=result.engine.name,
            log_messages=[self.convert(log) for log in result.log_messages],
        )



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Effect class.
A basic Effect has a fluent and an expression.
A condition can be added to make it a conditional effect.
"""


import unified_planning as up
from enum import Enum, auto
from typing import List, Callable, Dict


class EffectKind(Enum):
    ASSIGN = auto()
    INCREASE = auto()
    DECREASE = auto()


class Effect:
    def __init__(
        self,
        fluent: "up.model.fnode.FNode",
        value: "up.model.fnode.FNode",
        condition: "up.model.fnode.FNode",
        kind: EffectKind = EffectKind.ASSIGN,
    ):
        self._fluent = fluent
        self._value = value
        self._condition = condition
        self._kind = kind
        assert (
            fluent.environment == value.environment
            and value.environment == condition.environment
        ), "Effect expressions have different environment."

    def __repr__(self) -> str:
        s = []
        if self.is_conditional():
            s.append(f"if {str(self._condition)} then")
        s.append(f"{str(self._fluent)}")
        if self.is_assignment():
            s.append(":=")
        elif self.is_increase():
            s.append("+=")
        elif self.is_decrease():
            s.append("-=")
        s.append(f"{str(self._value)}")
        return " ".join(s)

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Effect):
            return (
                self._fluent == oth._fluent
                and self._value == oth._value
                and self._condition == oth._condition
                and self._kind == oth._kind
            )
        else:
            return False

    def __hash__(self) -> int:
        return (
            hash(self._fluent)
            + hash(self._value)
            + hash(self._condition)
            + hash(self._kind)
        )

    def clone(self):
        new_effect = Effect(self._fluent, self._value, self._condition, self._kind)
        return new_effect

    def is_conditional(self) -> bool:
        """Returns True if the Effect condition is not True."""
        return not self._condition.is_true()

    @property
    def fluent(self) -> "up.model.fnode.FNode":
        """Returns the Fluent that is modified by this effect."""
        return self._fluent

    @property
    def value(self) -> "up.model.fnode.FNode":
        """Returns the value given to the Fluent by this Effect."""
        return self._value

    def set_value(self, new_value: "up.model.fnode.FNode"):
        """Sets the value given to the Fluent by this Effect."""
        self._value = new_value

    @property
    def condition(self) -> "up.model.fnode.FNode":
        """Returns the condition required for this Effect to be applied."""
        return self._condition

    def set_condition(self, new_condition: "up.model.fnode.FNode"):
        """Sets the condition required for this Effect to be applied."""
        self._condition = new_condition

    @property
    def kind(self) -> EffectKind:
        """Returns the kind of this Effect."""
        return self._kind

    @property
    def environment(self) -> "up.environment.Environment":
        return self._fluent.environment

    def is_assignment(self) -> bool:
        """Returns True if the kind of this Effect is an assignment."""
        return self._kind == EffectKind.ASSIGN

    def is_increase(self) -> bool:
        """Returns True if the kind of this Effect is an increase."""
        return self._kind == EffectKind.INCREASE

    def is_decrease(self) -> bool:
        """Returns True if the kind of this Effect is a decrease."""
        return self._kind == EffectKind.DECREASE


class SimulatedEffect:
    """
    This class represents a simulated effect over a list of fluent expressions.
    The fluents parameters must be constants or action parameters.
    The callable function must return the result of the simulated effects applied
    in the given state for the specified fluent expressions.
    """

    def __init__(
        self,
        fluents: List["up.model.fnode.FNode"],
        function: Callable[
            [
                "up.model.problem.AbstractProblem",
                "up.model.state.ROState",
                Dict["up.model.parameter.Parameter", "up.model.fnode.FNode"],
            ],
            List["up.model.fnode.FNode"],
        ],
    ):
        for f in fluents:
            if not f.is_fluent_exp():
                raise up.exceptions.UPUsageError(
                    "Simulated effects can be defined on fluent expressions with constant parameters"
                )
            for c in f.args:
                if not (c.is_constant or c.is_parameter_exp()):
                    raise up.exceptions.UPUsageError(
                        "Simulated effects can be defined on fluent expressions with constant parameters"
                    )
        self._fluents = fluents
        self._function = function

    def __repr__(self) -> str:
        return f"{self._fluents} := simulated"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, SimulatedEffect):
            return self._fluents == oth._fluents and self._function == oth._function
        else:
            return False

    def __hash__(self) -> int:
        res = hash(self._function)
        for f in self._fluents:
            res += hash(f)
        return res

    @property
    def fluents(self) -> List["up.model.fnode.FNode"]:
        return self._fluents

    @property
    def function(
        self,
    ) -> Callable[
        [
            "up.model.problem.AbstractProblem",
            "up.model.state.ROState",
            Dict["up.model.parameter.Parameter", "up.model.fnode.FNode"],
        ],
        List["up.model.fnode.FNode"],
    ]:
        return self._function



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from unified_planning.model.expression import ConstantExpression
from unified_planning.exceptions import UPProblemDefinitionError, UPValueError
from typing import List, Dict, Union


class FluentsSetMixin:
    """
    This class is a mixin that contains a set of fluents with some related methods.

    NOTE: when this mixin is used in combination with other mixins that share some
    of the attributes (e.g. env, add_user_type_method, has_name_method), it is required
    to pass the very same arguments to the mixins constructors.
    """

    def __init__(
        self,
        env,
        add_user_type_method,
        has_name_method,
        initial_defaults: Dict["up.model.types.Type", "ConstantExpression"] = {},
    ):
        self._env = env
        self._add_user_type_method = add_user_type_method
        self._has_name_method = has_name_method
        self._fluents: List["up.model.fluent.Fluent"] = []
        self._fluents_defaults: Dict[
            "up.model.fluent.Fluent", "up.model.fnode.FNode"
        ] = {}
        self._initial_defaults: Dict["up.model.types.Type", "up.model.fnode.FNode"] = {}
        for k, v in initial_defaults.items():
            (v_exp,) = self.env.expression_manager.auto_promote(v)
            self._initial_defaults[k] = v_exp
        # The field initial default optionally associates a type to a default value. When a new fluent is
        # created with no explicit default, it will be associated with the initial-default of his type, if any.

    @property
    def env(self) -> "up.environment.Environment":
        """Returns the problem environment."""
        return self._env

    @property
    def fluents(self) -> List["up.model.fluent.Fluent"]:
        """Returns the fluents."""
        return self._fluents

    def fluent(self, name: str) -> "up.model.fluent.Fluent":
        """Returns the fluent with the given name."""
        for f in self._fluents:
            if f.name == name:
                return f
        raise UPValueError(f"Fluent of name: {name} is not defined!")

    def has_fluent(self, name: str) -> bool:
        """Returns true if the fluent with the given name is in the problem."""
        for f in self._fluents:
            if f.name == name:
                return True
        return False

    def add_fluents(self, fluents: List["up.model.fluent.Fluent"]):
        """Adds the given fluents."""
        for fluent in fluents:
            self.add_fluent(fluent)

    def add_fluent(
        self,
        fluent_or_name: Union["up.model.fluent.Fluent", str],
        typename: "up.model.types.Type" = None,
        *,
        default_initial_value: "ConstantExpression" = None,
        **kwargs: "up.model.types.Type",
    ) -> "up.model.fluent.Fluent":
        """Adds the given fluent to the problem.

        If the first parameter is not a Fluent, the parameters will be passed to the Fluent constructor to create it.

        :param fluent_or_name: Fluent instance or name of the fluent to be constructed
        :param typename: If only the name of the fluent is given, this is the fluent's type (passed to the Fluent constructor).
        :param default_initial_value: If provided, defines the default value taken in initial state by
                                      a state variable of this fluent that has no explicit value.
        :param kwargs: If only the name of the fluent is given, these are the fluent's parameters (passed to the Fluent constructor).
        :return: The fluent passed or constructed.

        Example
        --------
        >>> from unified_planning.shortcuts import *
        >>> problem = Problem()
        >>> location = UserType("Location")
        >>> at_loc = Fluent("at_loc", BoolType(), l=location)  # creates a new fluent
        >>> problem.add_fluent(at_loc)  # adds it to the problem
        bool at_loc[l=Location]
        >>> problem.add_fluent("connected", BoolType(), l1=location, l2=location)  # creates a new fluent and add it to the problem.
        bool connected[l1=Location, l2=Location]
        >>>
        """
        if isinstance(fluent_or_name, up.model.fluent.Fluent):
            assert len(kwargs) == 0 and typename is None
            fluent = fluent_or_name
            assert (
                fluent.environment == self._env
            ), "Fluent does not have the same environment of the problem"
        else:
            fluent = up.model.fluent.Fluent(
                fluent_or_name, typename, None, env=self.env, **kwargs
            )
        if self._has_name_method(fluent.name):
            raise UPProblemDefinitionError("Name " + fluent.name + " already defined!")
        self._fluents.append(fluent)
        if not default_initial_value is None:
            (v_exp,) = self.env.expression_manager.auto_promote(default_initial_value)
            self._fluents_defaults[fluent] = v_exp
        elif fluent.type in self._initial_defaults:
            self._fluents_defaults[fluent] = self._initial_defaults[fluent.type]
        if fluent.type.is_user_type():
            self._add_user_type_method(fluent.type)
        for param in fluent.signature:
            if param.type.is_user_type():
                self._add_user_type_method(param.type)

        return fluent

    @property
    def fluents_defaults(
        self,
    ) -> Dict["up.model.fluent.Fluent", "up.model.fnode.FNode"]:
        """Returns the problem's fluents defaults."""
        return self._fluents_defaults

    @property
    def initial_defaults(self) -> Dict["up.model.types.Type", "up.model.fnode.FNode"]:
        """Returns the problem's fluents defaults."""
        return self._initial_defaults



# Copyright 2022 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module defines the Task class.
A Task has a name and a signature that defines the types of its parameters.
"""

import unified_planning as up
from unified_planning.environment import get_env, Environment
from typing import List, OrderedDict, Optional, Union
from unified_planning.model.fnode import FNode
from unified_planning.model.action import Action
from unified_planning.model.timing import Timepoint, TimepointKind
from unified_planning.model.types import Type
from unified_planning.model.expression import Expression
from unified_planning.model.parameter import Parameter


class Task:
    """Represents an abstract task."""

    def __init__(
        self,
        name: str,
        _parameters: Optional[Union[OrderedDict[str, Type], List[Parameter]]] = None,
        _env: Environment = None,
        **kwargs: Type,
    ):
        self._env = get_env(_env)
        self._name = name
        self._parameters: List[Parameter] = []
        if _parameters is not None:
            assert len(kwargs) == 0
            if isinstance(_parameters, OrderedDict):
                for param_name, param_type in _parameters.items():
                    self._parameters.append(
                        up.model.parameter.Parameter(param_name, param_type, self._env)
                    )
            elif isinstance(_parameters, List):
                self._parameters = _parameters[:]
            else:
                raise NotImplementedError
        else:
            for param_name, param_type in kwargs.items():
                self._parameters.append(
                    up.model.parameter.Parameter(param_name, param_type, self._env)
                )

    def __repr__(self) -> str:
        sign = ""
        if len(self.parameters) > 0:
            sign_items = [f"{p.name}={str(p.type)}" for p in self.parameters]
            sign = f'[{", ".join(sign_items)}]'
        return f"{self.name}{sign}"

    def __eq__(self, oth: object) -> bool:
        if isinstance(oth, Task):
            return (
                self._name == oth._name
                and self._parameters == oth._parameters
                and self._env == oth._env
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash(self._name) + sum(map(hash, self._parameters))

    @property
    def name(self) -> str:
        """Returns the task's name."""
        return self._name

    @property
    def parameters(self) -> List[Parameter]:
        """Returns the task's parameters as a list."""
        return self._parameters

    def __call__(self, *args: Expression, ident: Optional[str] = None) -> "Subtask":
        """Returns a subtask with the given parameters."""
        return Subtask(self, *self._env.expression_manager.auto_promote(args))


# global counter to enable the creation of unique identifiers.
_task_id_counter = 0


class Subtask:
    def __init__(
        self,
        _task: Union[Action, Task],
        *args: Expression,
        ident: Optional[str] = None,
        _env: Environment = None,
    ):
        self._env = get_env(_env)
        self._task = _task
        self._ident: str
        if ident is not None:
            self._ident = ident
        else:
            # we have to create an unambiguous identifier as there might otherwise identical tasks
            global _task_id_counter
            _task_id_counter += 1
            self._ident = f"_t{_task_id_counter}"
        self._args = self._env.expression_manager.auto_promote(*args)
        assert len(self._args) == len(self._task.parameters)

    def __repr__(self):
        params = ", ".join([str(a) for a in self._args])
        return f"{self.identifier}: {self._task.name}({params})"

    def __eq__(self, other):
        if not isinstance(other, Subtask):
            return False
        return (
            self._env == other._env
            and self._ident == other._ident
            and self._task == other._task
            and self._args == other._args
        )

    def __hash__(self):
        return hash(self._ident) + hash(self._task) + sum(map(hash, self._args))

    @property
    def task(self) -> Union[Task, Action]:
        return self._task

    @property
    def parameters(self) -> List["FNode"]:
        return self._args

    @property
    def identifier(self) -> str:
        """Unique identifier of the subtask in its task network."""
        return self._ident

    @property
    def start(self) -> Timepoint:
        """Timepoint representing the task's end time."""
        return Timepoint(TimepointKind.START, container=self.identifier)

    @property
    def end(self) -> Timepoint:
        """Timepoint representing the task's end time."""
        return Timepoint(TimepointKind.END, container=self.identifier)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from fractions import Fraction
from typing import List, Optional, Union, cast

import unified_planning
import unified_planning.model
import unified_planning.model.walkers as walkers
from unified_planning.model.types import (
    _UserType as UT,
    _IntType as IT,
    _RealType as RT,
)

import tarski


class TarskiFormulaConverter(walkers.DagWalker):
    def __init__(
        self, language: "tarski.fol.FirstOrderLanguage", env, object_freshname: str
    ) -> None:
        walkers.DagWalker.__init__(self)
        self.object_freshname = object_freshname
        self.lang = language
        self.env = env

    def convert_formula(
        self, expression: "unified_planning.model.FNode"
    ) -> "tarski.syntax.formulas.Formula":
        return self.walk(expression)

    def walk_and(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        return tarski.syntax.land(*args, flat=True)

    def walk_or(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        return tarski.syntax.lor(*args, flat=True)

    def walk_not(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 1
        return tarski.syntax.neg(args[0])

    def walk_iff(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return tarski.syntax.equiv(args[0], args[1])

    def walk_implies(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return tarski.syntax.implies(args[0], args[1])

    def walk_exists(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 1
        variables = [
            self.lang.variable(
                v.name,
                self.lang.get_sort(
                    cast(UT, v.type).name
                    if cast(UT, v.type).name != "object"
                    else self.object_freshname
                ),
            )
            for v in expression.variables()
        ]
        return tarski.syntax.exists(*variables, args[0])

    def walk_forall(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 1
        variables = [
            self.lang.variable(
                v.name,
                self.lang.get_sort(
                    cast(UT, v.type).name
                    if cast(UT, v.type).name != "object"
                    else self.object_freshname
                ),
            )
            for v in expression.variables()
        ]
        return tarski.syntax.forall(*variables, args[0])

    def walk_equals(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return tarski.syntax.Atom(
            self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.EQ), args
        )

    def walk_le(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return tarski.syntax.Atom(
            self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.LE), args
        )

    def walk_lt(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return tarski.syntax.Atom(
            self.lang.get_predicate(tarski.syntax.BuiltinPredicateSymbol.LT), args
        )

    def walk_fluent_exp(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        tarski_fluent_rep = self.lang.get(expression.fluent().name)
        new_args = []
        for i, x in enumerate(expression.args):
            if x.is_int_constant():
                type = expression.fluent().signature[i].type
                typename = _type_name_added_to_language_if_needed(self.lang, type)
                constant = tarski.syntax.Constant(
                    x.int_constant_value(), self.lang.get_sort(typename)
                )
                new_args.append(constant)
            elif x.is_real_constant():
                raise unified_planning.exceptions.UPProblemDefinitionError(
                    "Fluents can not have reals into their signatures."
                )
            else:
                new_args.append(args[i])
        return tarski_fluent_rep(*new_args)

    def walk_plus(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        value = args[0]
        for a in args[1:]:
            value = value + a
        return value

    def walk_minus(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return args[0] - args[1]

    def walk_times(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        value = args[0]
        for a in args[1:]:
            value = value * a
        return value

    def walk_div(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 2
        return args[0] / args[1]

    def walk_bool_constant(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 0
        if expression.bool_constant_value():
            return tarski.syntax.top
        return tarski.syntax.bot

    def walk_int_constant(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 0
        return tarski.syntax.Constant(
            expression.int_constant_value(), self.lang.Integer
        )

    def walk_real_constant(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        assert len(args) == 0
        return tarski.syntax.Constant(expression.real_constant_value(), self.lang.Real)

    def walk_param_exp(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        type = expression.parameter().type
        if type.is_user_type():
            return self.lang.variable(
                expression.parameter().name,
                self.lang.get_sort(
                    cast(UT, type).name
                    if cast(UT, type).name != "object"
                    else self.object_freshname
                ),
            )
        else:
            return self.lang.variable(
                expression.parameter().name,
                self.lang.get_sort(
                    _type_name_added_to_language_if_needed(self.lang, type)
                ),
            )

    def walk_variable_exp(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        type = expression.variable().type
        if type.is_user_type():
            return self.lang.variable(
                expression.variable().name,
                self.lang.get_sort(
                    cast(UT, type).name
                    if cast(UT, type).name != "object"
                    else self.object_freshname
                ),
            )
        else:
            return self.lang.variable(
                expression.variable().name,
                self.lang.get_sort(
                    _type_name_added_to_language_if_needed(self.lang, type)
                ),
            )

    def walk_object_exp(
        self,
        expression: "unified_planning.model.FNode",
        args: List["tarski.syntax.formulas.Formula"],
    ) -> "tarski.syntax.formulas.Formula":
        return self.lang.get_constant(expression.object().name)


def convert_problem_to_tarski(
    problem: "unified_planning.model.Problem",
) -> "tarski.fstrips.problem.Problem":
    """Converts a problem in the unified_planning.model.Problem representation in the equivalent
    tarski.fstrips.Problem representation."""
    features: List[str] = []
    kind = problem.kind
    if kind.has_equality():
        features.append("equality")
    if (
        kind.has_continuous_numbers()
        or kind.has_discrete_numbers()
        or kind.has_numeric_fluents()
    ):
        features.append("arithmetic")
    object_freshname = "object_"
    while problem.has_type(object_freshname):
        object_freshname = object_freshname + "_"
    # creating tarski language
    lang = tarski.fstrips.language(f"{problem.name}_lang", features)
    for ut in problem.user_types:  # adding user_types to the language
        assert isinstance(ut, UT)
        if ut.father is not None:
            lang.sort(
                ut.name if ut.name != "object" else object_freshname,
                cast(UT, ut.father).name
                if cast(UT, ut.father).name != "object"
                else object_freshname,
            )
        else:
            lang.sort(ut.name if ut.name != "object" else object_freshname)
    for fluent in problem.fluents:  # adding fluents to the language
        signature = []
        for param in fluent.signature:
            if param.type.is_user_type():
                signature.append(
                    lang.get_sort(
                        cast(UT, param.type).name
                        if cast(UT, param.type).name != "object"
                        else object_freshname
                    )
                )
            else:
                # typename will be the name that this type has in the tarski language
                typename = _type_name_added_to_language_if_needed(lang, param.type)
                signature.append(lang.get_sort(typename))
        if fluent.type.is_bool_type():
            lang.predicate(fluent.name, *signature)
        else:
            typename = _type_name_added_to_language_if_needed(lang, fluent.type)
            lang.function(
                fluent.name,
                *signature,
                lang.get_sort(typename if typename != "object" else object_freshname),
            )
    for o in problem.all_objects:  # adding objects to the language
        lang.constant(
            o.name,
            lang.get_sort(
                cast(UT, o.type).name
                if cast(UT, o.type).name != "object"
                else object_freshname
            ),
        )
    # creating tarski problem
    em = problem.env.expression_manager
    tfc = TarskiFormulaConverter(
        language=lang, env=problem.env, object_freshname=object_freshname
    )
    new_problem = tarski.fstrips.problem.create_fstrips_problem(
        lang, problem.name, f"{problem.name}_domain"
    )
    for action in problem.actions:
        if not isinstance(action, unified_planning.model.InstantaneousAction):
            raise unified_planning.exceptions.UPProblemDefinitionError(
                "Tarski supports only Instantaneous Actions."
            )
        parameters = []
        for p in action.parameters:
            if p.type.is_user_type():
                parameters.append(
                    lang.variable(
                        p.name,
                        lang.get_sort(
                            cast(UT, p.type).name
                            if cast(UT, p.type).name != "object"
                            else object_freshname
                        ),
                    )
                )
            else:
                parameters.append(
                    lang.variable(
                        p.name,
                        lang.get_sort(
                            _type_name_added_to_language_if_needed(lang, p.type)
                        ),
                    )
                )
        # add action to the problem
        new_problem.action(
            action.name,
            parameters,
            precondition=tfc.convert_formula(em.And(action.preconditions)),
            effects=[_convert_effect(e, tfc, em) for e in action.effects],
        )
    for fluent_exp, value_exp in problem.initial_values.items():
        if value_exp.is_bool_constant():
            if value_exp.constant_value():
                parameters = []
                for a in fluent_exp.args:
                    parameters.append(tfc.convert_formula(a))
                new_problem.init.add(
                    lang.get_predicate(fluent_exp.fluent().name), *parameters
                )
        else:
            value: Optional[
                Union[Fraction, int, "tarski.syntax.formulas.Formula"]
            ] = None
            if value_exp.is_int_constant():
                value = value_exp.int_constant_value()
            elif value_exp.is_real_constant():
                value = value_exp.real_constant_value()
            else:
                value = tfc.convert_formula(value_exp)
            new_problem.init.set(tfc.convert_formula(fluent_exp), value)
    new_problem.goal = tfc.convert_formula(em.And(problem.goals))
    return new_problem


def _type_name_added_to_language_if_needed(
    lang: "tarski.FirstOrderLanguage", type: "unified_planning.model.Type"
) -> str:
    typename = ""
    if type.is_user_type():
        typename = cast(UT, type).name
    else:
        typename = str(type).replace(" ", "")
    if type.is_int_type() and (
        cast(IT, type).lower_bound is None or cast(IT, type).upper_bound is None
    ):
        type = cast(IT, type)
        if type.lower_bound is None and type.upper_bound is None:
            typename = "Integer"
        elif type.lower_bound == 0:
            assert (
                type.upper_bound is None
            )  # must be true, otherwise branch would be skipped
            typename = "Natural"
        else:
            raise unified_planning.exceptions.UPProblemDefinitionError(
                "Int type with just one bound is not accepted by tarski"
            )
    elif type.is_real_type() and (
        cast(RT, type).lower_bound is None or cast(RT, type).upper_bound is None
    ):
        type = cast(RT, type)
        if type.lower_bound is None and type.upper_bound is None:
            typename = "Real"
        else:
            raise unified_planning.exceptions.UPProblemDefinitionError(
                "Real type with just one bound is not accepted by tarski"
            )
    if not lang.has_sort(typename):
        # the type is not in the language, therefore it must be added
        if type.is_int_type():
            type = cast(IT, type)
            lang.interval(typename, lang.Integer, type.lower_bound, type.upper_bound)
        elif type.is_real_type():
            type = cast(RT, type)
            lang.interval(typename, lang.Real, type.lower_bound, type.upper_bound)
        else:
            raise NotImplementedError
    return typename


def _convert_effect(
    effect: "unified_planning.model.Effect",
    tfc: TarskiFormulaConverter,
    em: "unified_planning.model.ExpressionManager",
) -> "tarski.fstrips.SingleEffect":
    condition = tfc.convert_formula(effect.condition)
    predicate = tfc.convert_formula(effect.fluent)
    if effect.value.is_bool_constant():
        if effect.value.constant_value():
            return tarski.fstrips.AddEffect(predicate, condition)
        else:
            return tarski.fstrips.DelEffect(predicate, condition)
    else:
        if effect.kind == unified_planning.model.EffectKind.ASSIGN:
            value = tfc.convert_formula(effect.value)
        elif effect.kind == unified_planning.model.EffectKind.INCREASE:
            value = tfc.convert_formula(em.Plus(effect.fluent, effect.value))
        elif effect.kind == unified_planning.model.EffectKind.DECREASE:
            value = tfc.convert_formula(em.Minus(effect.fluent, effect.value))
        else:
            raise NotImplementedError
        return tarski.fstrips.FunctionalEffect(predicate, value, condition)



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unified_planning as up
from fractions import Fraction
from typing import Dict, Optional, Union


class PlanQualityMetric:
    """This is the base class of any metric for plan quality"""

    pass


class MinimizeActionCosts(PlanQualityMetric):
    def __init__(
        self,
        costs: Dict["up.model.Action", "up.model.FNode"],
        default: Optional["up.model.FNode"] = None,
    ):
        self.costs = costs
        self.default = default

    def __repr__(self):
        costs = {a.name: c for a, c in self.costs.items()}
        costs["default"] = self.default
        return f"minimize actions-cost: {costs}"

    def get_action_cost(self, action: "up.model.Action") -> Optional["up.model.FNode"]:
        return self.costs.get(action, self.default)


class MinimizeSequentialPlanLength(PlanQualityMetric):
    def __repr__(self):
        return "minimize sequential-plan-length"


class MinimizeMakespan(PlanQualityMetric):
    def __repr__(self):
        return "minimize makespan"


class MinimizeExpressionOnFinalState(PlanQualityMetric):
    def __init__(self, expression: "up.model.FNode"):
        self.expression = expression

    def __repr__(self):
        return f"minimize {self.expression}"


class MaximizeExpressionOnFinalState(PlanQualityMetric):
    def __init__(self, expression: "up.model.FNode"):
        self.expression = expression

    def __repr__(self):
        return f"maximize {self.expression}"


class Oversubscription(PlanQualityMetric):
    def __init__(self, goals: Dict["up.model.FNode", Union[Fraction, int]]):
        self.goals = goals

    def __repr__(self):
        return f"oversubscription planning goals: {self.goals}"



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import unified_planning.model.types
import unified_planning.environment
import unified_planning.model.walkers as walkers
from unified_planning.model.types import BOOL, TIME, _IntType, _RealType
from unified_planning.model.fnode import FNode
from unified_planning.model.operators import OperatorKind
from unified_planning.exceptions import UPTypeError
from typing import List, Optional


class TypeChecker(walkers.dag.DagWalker):
    def __init__(self, env: "unified_planning.environment.Environment"):
        walkers.dag.DagWalker.__init__(self)
        self.env = env

    def get_type(self, expression: FNode) -> "unified_planning.model.types.Type":
        """Returns the unified_planning.model.types type of the expression"""
        res = self.walk(expression)
        if res is None:
            raise UPTypeError(
                "The expression '%s' is not well-formed" % str(expression)
            )
        return res

    @walkers.handles(
        OperatorKind.AND,
        OperatorKind.OR,
        OperatorKind.NOT,
        OperatorKind.IMPLIES,
        OperatorKind.IFF,
        OperatorKind.EXISTS,
        OperatorKind.FORALL,
    )
    def walk_bool_to_bool(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> Optional["unified_planning.model.types.Type"]:
        assert expression is not None
        for x in args:
            if x is None or x != BOOL:
                return None
        return BOOL

    def walk_fluent_exp(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> Optional["unified_planning.model.types.Type"]:
        assert expression.is_fluent_exp()
        f = expression.fluent()
        if len(args) != len(f.signature):
            return None
        for (arg, param) in zip(args, f.signature):
            if not arg.is_compatible(param.type):
                return None
        return f.type

    def walk_param_exp(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> "unified_planning.model.types.Type":
        assert expression is not None
        assert len(args) == 0
        return expression.parameter().type

    def walk_variable_exp(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> "unified_planning.model.types.Type":
        assert expression is not None
        assert len(args) == 0
        return expression.variable().type

    def walk_object_exp(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> "unified_planning.model.types.Type":
        assert expression is not None
        assert len(args) == 0
        return expression.object().type

    def walk_timing_exp(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> "unified_planning.model.types.Type":
        assert expression is not None
        assert len(args) == 0
        return TIME

    @walkers.handles(OperatorKind.BOOL_CONSTANT)
    def walk_identity_bool(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> Optional["unified_planning.model.types.Type"]:
        assert expression is not None
        assert len(args) == 0
        return BOOL

    @walkers.handles(OperatorKind.REAL_CONSTANT)
    def walk_identity_real(self, expression, args):
        assert expression is not None
        assert len(args) == 0
        return self.env.type_manager.RealType(
            expression.constant_value(), expression.constant_value()
        )

    @walkers.handles(OperatorKind.INT_CONSTANT)
    def walk_identity_int(self, expression, args):
        assert expression is not None
        assert len(args) == 0
        return self.env.type_manager.IntType(
            expression.constant_value(), expression.constant_value()
        )

    def walk_plus(self, expression, args):
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        for x in args:
            if x.lower_bound is None:
                lower = -float("inf")
            elif lower is None:
                lower = x.lower_bound
            else:
                lower += x.lower_bound
            if x.upper_bound is None:
                upper = float("inf")
            elif upper is None:
                upper = x.upper_bound
            else:
                upper += x.upper_bound
        if lower == -float("inf"):
            lower = None
        if upper == float("inf"):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_minus(self, expression, args):
        assert len(args) == 2
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        left = args[0]
        right = args[1]
        left_lower = -float("inf") if left.lower_bound is None else left.lower_bound
        left_upper = float("inf") if left.upper_bound is None else left.upper_bound
        right_lower = -float("inf") if right.lower_bound is None else right.lower_bound
        right_upper = float("inf") if right.upper_bound is None else right.upper_bound
        lower = left_lower - right_upper
        upper = left_upper - right_lower
        if lower == -float("inf"):
            lower = None
        if upper == float("inf"):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_times(self, expression, args):
        has_real = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
        for x in args:
            l = -float("inf") if x.lower_bound is None else x.lower_bound
            u = float("inf") if x.upper_bound is None else x.upper_bound
            if lower is None:
                lower = l
                upper = u
            else:
                lower = min(lower * l, lower * u, upper * l, upper * u)
                upper = max(lower * l, lower * u, upper * l, upper * u)
        if lower == -float("inf"):
            lower = None
        if upper == float("inf"):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    def walk_div(self, expression, args):
        assert len(args) == 2
        has_real = False
        to_skip = False
        lower = None
        upper = None
        for x in args:
            if x is None or not (x.is_int_type() or x.is_real_type()):
                return None
            if x.is_real_type():
                has_real = True
            if x.lower_bound is None and x.upper_bound is None:
                to_skip = True
        left = args[0]
        right = args[1]
        if to_skip or right.lower_bound != right.upper_bound:
            pass
        else:
            left_lower = -float("inf") if left.lower_bound is None else left.lower_bound
            left_upper = float("inf") if left.upper_bound is None else left.upper_bound
            right = right.lower_bound
            lower = min(left_lower / right, left_upper / right)
            upper = max(left_lower / right, left_upper / right)
        if lower == -float("inf"):
            lower = None
        if upper == float("inf"):
            upper = None
        if has_real:
            return self.env.type_manager.RealType(lower, upper)
        else:
            return self.env.type_manager.IntType(lower, upper)

    @walkers.handles(OperatorKind.LE, OperatorKind.LT)
    def walk_math_relation(self, expression, args):
        for x in args:
            if x is None or not (
                x.is_int_type() or x.is_real_type() or x.is_time_type()
            ):
                return None
        return BOOL

    def walk_equals(
        self, expression: FNode, args: List["unified_planning.model.types.Type"]
    ) -> Optional["unified_planning.model.types.Type"]:
        t = args[0]
        if t is None:
            return None

        if t.is_bool_type():
            raise UPTypeError(
                "The expression '%s' is not well-formed."
                "Equality operator is not supported for Boolean"
                " terms. Use Iff instead." % str(expression)
            )
        for x in args:
            if x is None:
                return None
            elif t.is_user_type() and t != x:
                return None
            elif (t.is_int_type() or t.is_real_type()) and not (
                x.is_int_type() or x.is_real_type()
            ):
                return None
        return BOOL



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# type: ignore[valid-type]
import fractions
from itertools import product
from typing import List

import unified_planning.grpc.generated.unified_planning_pb2 as proto
from unified_planning import model
import unified_planning.model.htn
import unified_planning.model.walkers as walkers
from unified_planning.model.types import domain_size, domain_item
from unified_planning.exceptions import UPException
from unified_planning.grpc.converter import Converter, handles
from unified_planning.model.operators import (
    BOOL_OPERATORS,
    IRA_OPERATORS,
    RELATIONS,
    OperatorKind,
)
from unified_planning.model.timing import TimepointKind


def map_operator(op: int) -> str:
    if op == OperatorKind.PLUS:
        return "up:plus"
    elif op == OperatorKind.MINUS:
        return "up:minus"
    elif op == OperatorKind.TIMES:
        return "up:times"
    elif op == OperatorKind.DIV:
        return "up:div"
    elif op == OperatorKind.LE:
        return "up:le"
    elif op == OperatorKind.LT:
        return "up:lt"
    elif op == OperatorKind.EQUALS:
        return "up:equals"
    elif op == OperatorKind.AND:
        return "up:and"
    elif op == OperatorKind.OR:
        return "up:or"
    elif op == OperatorKind.NOT:
        return "up:not"
    elif op == OperatorKind.IMPLIES:
        return "up:implies"
    elif op == OperatorKind.IFF:
        return "up:iff"
    elif op == OperatorKind.EXISTS:
        return "up:exists"
    elif op == OperatorKind.FORALL:
        return "up:forall"
    raise ValueError(f"Unknown operator `{op}`")


def proto_type(tpe: model.Type) -> str:
    if tpe.is_bool_type():
        return "up:bool"
    elif tpe.is_time_type():
        return "up:time"
    elif tpe.is_int_type() or tpe.is_real_type():
        return f"up:{tpe}"
    elif isinstance(tpe, model.types._UserType):
        return str(tpe.name)


class FNode2Protobuf(walkers.DagWalker):
    def __init__(self, protobuf_writer):
        super().__init__()
        self._protobuf_writer = protobuf_writer

    def convert(self, expression: model.FNode) -> proto.Expression:
        return self.walk(expression)

    def walk_bool_constant(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(boolean=expression.bool_constant_value()),
            list=[],
            kind=proto.ExpressionKind.Value("CONSTANT"),
            type="up:bool",
        )

    def walk_int_constant(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(int=expression.int_constant_value()),
            list=[],
            kind=proto.ExpressionKind.Value("CONSTANT"),
            type="up:integer",
        )

    def walk_real_constant(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(
                real=self._protobuf_writer.convert(expression.real_constant_value())
            ),
            list=[],
            kind=proto.ExpressionKind.Value("CONSTANT"),
            type="up:real",
        )

    def walk_param_exp(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(symbol=expression.parameter().name),
            list=[],
            kind=proto.ExpressionKind.Value("PARAMETER"),
            type=proto_type(expression.parameter().type),
        )

    def walk_variable_exp(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(symbol=expression.variable().name),
            list=[],
            kind=proto.ExpressionKind.Value("VARIABLE"),
            type=proto_type(expression.variable().type),
        )

    def walk_object_exp(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        return proto.Expression(
            atom=proto.Atom(symbol=expression.object().name),
            list=[],
            kind=proto.ExpressionKind.Value("CONSTANT"),
            type=proto_type(expression.object().type),
        )

    def walk_timing_exp(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        timing = expression.timing()
        tp = timing.timepoint
        if timing.timepoint.container is not None:
            args = [
                proto.Expression(
                    atom=proto.Atom(symbol=timing.timepoint.container),
                    type="up:container",
                    kind=proto.ExpressionKind.Value("CONTAINER_ID"),
                )
            ]
        else:
            args = []
        if tp.kind == TimepointKind.GLOBAL_START:
            fn = "up:global_start"
        elif tp.kind == TimepointKind.GLOBAL_END:
            fn = "up:global_end"
        elif tp.kind == TimepointKind.START:
            fn = "up:start"
        elif tp.kind == TimepointKind.END:
            fn = "up:end"
        else:
            raise ValueError(f"Unknown timepoint kind: {tp.kind}")
        fn_exp = proto.Expression(
            atom=proto.Atom(symbol=fn),
            kind=proto.ExpressionKind.Value("FUNCTION_SYMBOL"),
        )
        tp_exp = proto.Expression(
            list=[fn_exp] + args,
            type="up:time",
            kind=proto.ExpressionKind.Value("FUNCTION_APPLICATION"),
        )
        assert timing.delay == 0
        return tp_exp

    def walk_fluent_exp(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        sub_list = []
        sub_list.append(
            proto.Expression(
                atom=proto.Atom(symbol=expression.fluent().name),
                kind=proto.ExpressionKind.Value("FLUENT_SYMBOL"),
                type=proto_type(expression.fluent().type),
            )
        )
        sub_list.extend(args)
        return proto.Expression(
            atom=None,
            list=sub_list,
            kind=proto.ExpressionKind.Value("STATE_VARIABLE"),
            type=proto_type(expression.fluent().type),
        )

    @walkers.handles(BOOL_OPERATORS.union(IRA_OPERATORS).union(RELATIONS))
    def walk_operator(
        self, expression: model.FNode, args: List[proto.Expression]
    ) -> proto.Expression:
        sub_list = []
        sub_list.append(
            proto.Expression(
                atom=proto.Atom(symbol=map_operator(expression.node_type)),
                list=[],
                kind=proto.ExpressionKind.Value("FUNCTION_SYMBOL"),
                type="up:operator",
            )
        )
        # forall/exists: add the declared variables from the payload to the beginning of the parameter list.
        if expression.is_exists() or expression.is_forall():
            sub_list.extend(
                [self._protobuf_writer.convert(p) for p in expression.variables()]
            )

        sub_list.extend(args)
        return proto.Expression(
            atom=None,
            list=sub_list,
            kind=proto.ExpressionKind.Value("FUNCTION_APPLICATION"),
            type="",
        )


def map_feature(feature: str) -> proto.Feature:
    pb_feature = proto.Feature.Value(feature)
    if pb_feature is None:
        raise ValueError(f"Cannot convert feature to protobuf {feature}")
    return pb_feature


class ProtobufWriter(Converter):
    def __init__(self):
        super().__init__()
        self._fnode2proto = FNode2Protobuf(self)

    @handles(model.Fluent)
    def _convert_fluent(
        self, fluent: model.Fluent, problem: model.Problem
    ) -> proto.Fluent:
        name = fluent.name
        sig = [self.convert(t) for t in fluent.signature]
        return proto.Fluent(
            name=name,
            value_type=proto_type(fluent.type),
            parameters=sig,
            default_value=self.convert(problem.fluents_defaults[fluent])
            if fluent in problem.fluents_defaults
            else None,
        )

    @handles(model.Object)
    def _convert_object(self, obj: model.Object) -> proto.ObjectDeclaration:
        return proto.ObjectDeclaration(name=obj.name, type=proto_type(obj.type))

    @handles(model.FNode)
    def _convert_fnode(self, exp: model.FNode) -> proto.Expression:
        return self._fnode2proto.convert(exp)

    @handles(model.types._BoolType)
    def _convert_bool_type(self, tpe: model.types._BoolType) -> proto.TypeDeclaration:
        return proto.TypeDeclaration(type_name=proto_type(tpe))

    @handles(model.types._UserType)
    def _convert_user_type(self, t: model.types._UserType) -> proto.TypeDeclaration:
        return proto.TypeDeclaration(
            type_name=proto_type(t),
            parent_type="" if t.father is None else proto_type(t.father),
        )

    @handles(model.types._IntType)
    def _convert_integer_type(self, t: model.types._IntType) -> proto.TypeDeclaration:
        return proto.TypeDeclaration(type_name=proto_type(t))

    @handles(model.types._RealType)
    def _convert_real(self, t: model.types._RealType) -> proto.TypeDeclaration:
        return proto.TypeDeclaration(type_name=proto_type(t))

    @handles(model.Effect)
    def _convert_effect(self, effect: model.Effect) -> proto.Effect:
        kind = None
        if effect.is_assignment():
            kind = proto.EffectExpression.EffectKind.Value("ASSIGN")
        elif effect.is_increase():
            kind = proto.EffectExpression.EffectKind.Value("INCREASE")
        elif effect.is_decrease():
            kind = proto.EffectExpression.EffectKind.Value("DECREASE")
        else:
            raise ValueError(f"Unsupported effect: {effect}")

        return proto.EffectExpression(
            kind=kind,
            fluent=self.convert(effect.fluent),
            value=self.convert(effect.value),
            condition=self.convert(effect.condition),
        )

    @handles(model.InstantaneousAction)
    def _convert_instantaneous_action(
        self, a: model.InstantaneousAction
    ) -> proto.Action:
        effects = []
        conditions = []

        for cond in a.preconditions:
            conditions.append(
                proto.Condition(
                    cond=self.convert(cond),
                    span=None,
                )
            )

        for eff in a.effects:
            effects.append(proto.Effect(effect=self.convert(eff), occurrence_time=None))

        return proto.Action(
            name=a.name,
            parameters=[self.convert(p) for p in a.parameters],
            duration=None,
            conditions=conditions,
            effects=effects,
        )

    @handles(model.DurativeAction)
    def _convert_durative_action(self, a: model.DurativeAction) -> proto.Action:
        effects = []
        conditions = []

        for span, cond in a.conditions.items():
            span = self.convert(span)
            for c in cond:
                conditions.append(
                    proto.Condition(
                        cond=self.convert(c),
                        span=span,
                    )
                )
        for ot, eff in a.effects.items():
            ot = self.convert(ot)
            for e in eff:
                effects.append(
                    proto.Effect(
                        effect=self.convert(e),
                        occurrence_time=ot,
                    )
                )

        return proto.Action(
            name=a.name,
            parameters=[self.convert(p) for p in a.parameters],
            duration=self.convert(a.duration),
            conditions=conditions,
            effects=effects,
        )

    @handles(model.timing.Timepoint)
    def _convert_timepoint(self, tp: model.timing.Timepoint) -> proto.Timepoint:
        if tp.kind == TimepointKind.START:
            kind = proto.Timepoint.TimepointKind.Value("START")
        elif tp.kind == TimepointKind.END:
            kind = proto.Timepoint.TimepointKind.Value("END")
        elif tp.kind == TimepointKind.GLOBAL_START:
            kind = proto.Timepoint.TimepointKind.Value("GLOBAL_START")
        elif tp.kind == TimepointKind.GLOBAL_END:
            kind = proto.Timepoint.TimepointKind.Value("GLOBAL_END")
        return proto.Timepoint(kind=kind, container_id=tp.container)

    @handles(model.Timing)
    def _convert_timing(self, timing: model.Timing) -> proto.Timing:
        return proto.Timing(
            timepoint=self.convert(timing._timepoint),
            delay=self.convert(fractions.Fraction(timing.delay)),
        )

    @handles(fractions.Fraction)
    def _convert_fraction(self, fraction: fractions.Fraction) -> float:
        return proto.Real(
            numerator=fraction.numerator, denominator=fraction.denominator
        )

    @handles(model.timing.Interval)
    def _convert_interval(self, interval: model.timing.Interval) -> proto.Interval:
        return proto.Interval(
            is_left_open=interval.is_left_open(),
            lower=self.convert(interval.lower()),
            is_right_open=interval.is_right_open(),
            upper=self.convert(interval.lower()),
        )

    @handles(model.TimeInterval)
    def _convert_time_interval(
        self, interval: model.TimeInterval
    ) -> proto.TimeInterval:
        return proto.TimeInterval(
            is_left_open=interval.is_left_open(),
            lower=self.convert(interval.lower),
            is_right_open=interval.is_right_open(),
            upper=self.convert(interval.upper),
        )

    @handles(model.DurationInterval)
    def _convert_duration_interval(
        self, interval: model.DurationInterval
    ) -> proto.Interval:
        return proto.Duration(
            controllable_in_bounds=proto.Interval(
                is_left_open=interval.is_left_open(),
                lower=self.convert(interval.lower),
                is_right_open=interval.is_right_open(),
                upper=self.convert(interval.upper),
            )
        )

    @handles(model.htn.Task)
    def _convert_abstract_task(
        self, task: model.htn.Task
    ) -> proto.AbstractTaskDeclaration:
        return proto.AbstractTaskDeclaration(
            name=task.name, parameters=[self.convert(p) for p in task.parameters]
        )

    @handles(model.htn.ParameterizedTask)
    def _convert_parameterized_task(
        self, task: model.htn.ParameterizedTask
    ) -> proto.Task:
        parameters = []
        for p in task.parameters:
            parameters.append(
                proto.Expression(
                    atom=proto.Atom(symbol=p.name),
                    list=[],
                    kind=proto.ExpressionKind.Value("PARAMETER"),
                    type=proto_type(p.type),
                )
            )
        return proto.Task(id="", task_name=task.task.name, parameters=parameters)

    @handles(model.htn.Subtask)
    def _convert_subtask(self, subtask: model.htn.Subtask) -> proto.Task:
        return proto.Task(
            id=subtask.identifier,
            task_name=subtask.task.name,
            parameters=[self.convert(p) for p in subtask.parameters],
        )

    @handles(model.htn.Method)
    def _convert_method(self, method: model.htn.Method) -> proto.Method:
        return proto.Method(
            name=method.name,
            parameters=[self.convert(p) for p in method.parameters],
            achieved_task=self.convert(method.achieved_task),
            subtasks=[self.convert(st) for st in method.subtasks],
            constraints=[self.convert(c) for c in method.constraints],
            conditions=[
                proto.Condition(cond=self.convert(c)) for c in method.preconditions
            ],
        )

    @handles(model.htn.TaskNetwork)
    def _convert_task_network(self, tn: model.htn.TaskNetwork) -> proto.TaskNetwork:
        return proto.TaskNetwork(
            variables=[self.convert(v) for v in tn.variables],
            subtasks=[self.convert(st) for st in tn.subtasks],
            constraints=[self.convert(c) for c in tn.constraints],
        )

    def build_hierarchy(
        self, problem: model.htn.HierarchicalProblem
    ) -> proto.Hierarchy:
        return proto.Hierarchy(
            initial_task_network=self.convert(problem.task_network),
            abstract_tasks=[self.convert(t) for t in problem.tasks],
            methods=[self.convert(m) for m in problem.methods],
        )

    @handles(model.Problem, model.htn.HierarchicalProblem)
    def _convert_problem(self, problem: model.Problem) -> proto.Problem:
        goals = [proto.Goal(goal=self.convert(g)) for g in problem.goals]
        for (t, gs) in problem.timed_goals:
            goals += [
                proto.Goal(goal=self.convert(g), timing=self.convert(t)) for g in gs
            ]

        problem_name = str(problem.name) if problem.name is not None else ""
        hierarchy = None
        if isinstance(problem, model.htn.HierarchicalProblem):
            hierarchy = self.build_hierarchy(problem)

        return proto.Problem(
            domain_name=problem_name + "_domain",
            problem_name=problem_name,
            types=[self.convert(t) for t in problem.user_types],
            fluents=[self.convert(f, problem) for f in problem.fluents],
            objects=[self.convert(o) for o in problem.all_objects],
            actions=[self.convert(a) for a in problem.actions],
            initial_state=[
                proto.Assignment(fluent=self.convert(x), value=self.convert(v))
                for (x, v) in problem.initial_values.items()
            ],
            timed_effects=[self.convert(e) for e in problem.timed_effects],
            goals=goals,
            features=[map_feature(feature) for feature in problem.kind.features],
            metrics=[self.convert(m) for m in problem.quality_metrics],
            hierarchy=hierarchy,
        )

    @handles(model.metrics.MinimizeActionCosts)
    def _convert_minimize_action_costs(
        self, metric: model.metrics.MinimizeActionCosts
    ) -> proto.Metric:
        action_costs = {}
        for action, cost in metric.costs.items():
            action_costs[action.name] = self.convert(cost)

        return proto.Metric(
            kind=proto.Metric.MINIMIZE_ACTION_COSTS,
            action_costs=action_costs,
            default_action_cost=self.convert(metric.default)
            if metric.default is not None
            else None,
        )

    @handles(model.metrics.MinimizeSequentialPlanLength)
    def _convert_minimize_sequential_plan_length(self, _) -> proto.Metric:
        return proto.Metric(
            kind=proto.Metric.MINIMIZE_SEQUENTIAL_PLAN_LENGTH,
        )

    @handles(model.metrics.MinimizeMakespan)
    def _convert_minimize_makespan(self, _) -> proto.Metric:
        return proto.Metric(
            kind=proto.Metric.MINIMIZE_MAKESPAN,
        )

    @handles(model.metrics.MinimizeExpressionOnFinalState)
    def _convert_minimize_expression_on_final_state(
        self, metric: model.metrics.MinimizeExpressionOnFinalState
    ) -> proto.Metric:
        return proto.Metric(
            kind=proto.Metric.MINIMIZE_EXPRESSION_ON_FINAL_STATE,
            expression=self.convert(metric.expression),
        )

    @handles(model.metrics.MaximizeExpressionOnFinalState)
    def _convert_maximize_expression_on_final_state(
        self, metric: model.metrics.MaximizeExpressionOnFinalState
    ) -> proto.Metric:
        return proto.Metric(
            kind=proto.Metric.MAXIMIZE_EXPRESSION_ON_FINAL_STATE,
            expression=self.convert(metric.expression),
        )

    @handles(model.metrics.Oversubscription)
    def _convert_oversubscription_metric(
        self, metric: model.metrics.Oversubscription
    ) -> proto.Metric:
        goals = []
        for g, c in metric.goals.items():
            goals.append(
                proto.GoalWithCost(
                    goal=self.convert(g), cost=self.convert(fractions.Fraction(c))
                )
            )
        return proto.Metric(
            kind=proto.Metric.OVERSUBSCRIPTION,
            goals=goals,
        )

    @handles(model.Parameter)
    def _convert_action_parameter(self, p: model.Parameter) -> proto.Parameter:
        return proto.Parameter(name=p.name, type=proto_type(p.type))

    @handles(model.Variable)
    def _convert_expression_variable(
        self, variable: model.Variable
    ) -> proto.Expression:
        # a variable declaration (in forall/exists) is converted directly as an expression
        return proto.Expression(
            atom=proto.Atom(symbol=variable.name),
            list=[],
            kind=proto.ExpressionKind.Value("VARIABLE"),
            type=proto_type(variable.type),
        )

    @handles(unified_planning.plans.ActionInstance)
    def _convert_action_instance(
        self, a: unified_planning.plans.ActionInstance, start_time=None, end_time=None
    ) -> proto.ActionInstance:
        parameters = []
        for param in a.actual_parameters:
            # The parameters are atoms
            parameters.append(self.convert(param).atom)

        return proto.ActionInstance(
            action_name=a.action.name,
            parameters=parameters,
            start_time=start_time,
            end_time=end_time,
        )

    @handles(str)
    def _convert_str_atom(self, s: str) -> proto.Atom:
        return proto.Atom(symbol=s)

    @handles(unified_planning.plans.SequentialPlan)
    def _convert_sequential_plan(
        self, plan: unified_planning.plans.SequentialPlan
    ) -> proto.Plan:
        return proto.Plan(actions=[self.convert(a) for a in plan.actions])

    @handles(unified_planning.plans.TimeTriggeredPlan)
    def _convert_time_triggered_plan(
        self, plan: unified_planning.plans.TimeTriggeredPlan
    ) -> proto.Plan:
        action_instances = []

        for a in plan.timed_actions:
            start_time = self.convert(a[0])
            end_time = self.convert(a[0] + a[2])
            instance = self._convert_action_instance(
                a[1], start_time=start_time, end_time=end_time
            )
            action_instances.append(instance)

        return proto.Plan(actions=action_instances)

    @handles(unified_planning.engines.PlanGenerationResult)
    def _convert_plan_generation_result(
        self, result: unified_planning.engines.PlanGenerationResult
    ) -> proto.PlanGenerationResult:
        log_messages = None
        if result.log_messages is not None:
            log_messages = [self.convert(log) for log in result.log_messages]

        return proto.PlanGenerationResult(
            status=self.convert(result.status),
            plan=self.convert(result.plan),
            engine=proto.Engine(name=result.engine_name),
            metrics=result.metrics,
            log_messages=log_messages,
        )

    @handles(unified_planning.engines.PlanGenerationResultStatus)
    def _convert_plan_generation_status(
        self, status: unified_planning.engines.PlanGenerationResultStatus
    ) -> proto.PlanGenerationResult.Status:
        if (
            status
            == unified_planning.engines.PlanGenerationResultStatus.SOLVED_SATISFICING
        ):
            return proto.PlanGenerationResult.Status.Value("SOLVED_SATISFICING")

        elif (
            status
            == unified_planning.engines.PlanGenerationResultStatus.SOLVED_OPTIMALLY
        ):
            return proto.PlanGenerationResult.Status.Value("SOLVED_OPTIMALLY")
        elif (
            status
            == unified_planning.engines.PlanGenerationResultStatus.UNSOLVABLE_PROVEN
        ):
            return proto.PlanGenerationResult.Status.Value("UNSOLVABLE_PROVEN")
        elif (
            status
            == unified_planning.engines.PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY
        ):
            return proto.PlanGenerationResult.Status.Value("UNSOLVABLE_INCOMPLETELY")
        elif status == unified_planning.engines.PlanGenerationResultStatus.TIMEOUT:
            return proto.PlanGenerationResult.Status.Value("TIMEOUT")
        elif status == unified_planning.engines.PlanGenerationResultStatus.MEMOUT:
            return proto.PlanGenerationResult.Status.Value("MEMOUT")
        elif (
            status == unified_planning.engines.PlanGenerationResultStatus.INTERNAL_ERROR
        ):
            return proto.PlanGenerationResult.Status.Value("INTERNAL_ERROR")
        elif (
            status
            == unified_planning.engines.PlanGenerationResultStatus.UNSUPPORTED_PROBLEM
        ):

            return proto.PlanGenerationResult.Status.Value("UNSUPPORTED_PROBLEM")
        elif status == unified_planning.engines.PlanGenerationResultStatus.INTERMEDIATE:
            return proto.PlanGenerationResult.Status.Value("INTERMEDIATE")
        else:
            raise ValueError("Unknown status: {}".format(status))

    @handles(unified_planning.engines.LogMessage)
    def _convert_log_messages(
        self, log: unified_planning.engines.LogMessage
    ) -> proto.LogMessage:
        if log.level == unified_planning.engines.LogLevel.INFO:
            level = proto.LogMessage.LogLevel.Value("INFO")
        elif log.level == unified_planning.engines.LogLevel.WARNING:
            level = proto.LogMessage.LogLevel.Value("WARNING")
        elif log.level == unified_planning.engines.LogLevel.ERROR:
            level = proto.LogMessage.LogLevel.Value("ERROR")
        elif log.level == unified_planning.engines.LogLevel.DEBUG:
            level = proto.LogMessage.LogLevel.Value("DEBUG")
        else:
            raise UPException(f"Unknown log level: {log.level}")

        return proto.LogMessage(
            level=level,
            message=str(log.message),
        )

    @handles(unified_planning.engines.CompilerResult)
    def _convert_compiler_result(
        self, result: unified_planning.engines.CompilerResult
    ) -> proto.CompilerResult:
        map: Dict[str, proto.ActionInstance] = {}
        log_messages = result.log_messages
        if log_messages is None:
            log_messages = []
        if result.map_back_action_instance is not None:
            for compiled_action in result.problem.actions:
                type_list = [param.type for param in compiled_action.parameters]
                if len(type_list) == 0:
                    ai = unified_planning.plans.ActionInstance(compiled_action)
                    map[str(ai)] = self.convert(result.map_back_action_instance(ai))
                    continue
                ground_size = 1
                domain_sizes = []
                for t in type_list:
                    ds = domain_size(result.problem, t)
                    domain_sizes.append(ds)
                    ground_size *= ds
                items_list: List[List[FNode]] = []
                for size, type in zip(domain_sizes, type_list):
                    items_list.append(
                        [domain_item(result.problem, type, j) for j in range(size)]
                    )
                grounded_params_list = product(*items_list)
                for grounded_params in grounded_params_list:
                    ai = unified_planning.plans.ActionInstance(
                        compiled_action, tuple(grounded_params)
                    )
                    map[str(ai)] = self.convert(result.map_back_action_instance(ai))
        return proto.CompilerResult(
            problem=self.convert(result.problem),
            map_back_plan=map,
            log_messages=[self.convert(log) for log in log_messages],
            engine=proto.Engine(name=result.engine_name),
        )

    @handles(unified_planning.engines.ValidationResult)
    def _convert_validation_result(
        self, result: unified_planning.engines.ValidationResult
    ) -> proto.ValidationResult:
        return proto.ValidationResult(
            status=self.convert(result.status),
            log_messages=[self.convert(log) for log in result.log_messages],
            engine=proto.Engine(name=result.engine_name),
        )

    @handles(unified_planning.engines.ValidationResultStatus)
    def _convert_validation_result_status(
        self, status: unified_planning.engines.ValidationResultStatus
    ) -> proto.ValidationResult.ValidationResultStatus:
        if status == unified_planning.engines.ValidationResultStatus.VALID:
            return proto.ValidationResult.ValidationResultStatus.Value("VALID")
        elif status == unified_planning.engines.ValidationResultStatus.INVALID:
            return proto.ValidationResult.ValidationResultStatus.Value("INVALID")
        else:
            raise UPException(f"Unknown result status: {status}")



# Copyright 2021 AIPlan4EU project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from typing import Dict, List, Optional
import unified_planning as up
from unified_planning.model.fnode import FNode
from unified_planning.model.expression import Expression
from unified_planning.exceptions import UPProblemDefinitionError
from unified_planning.model.walkers.quantifier_simplifier import QuantifierSimplifier


class StateEvaluator(QuantifierSimplifier):
    """Same to the unified_planning.QuantifierSimplifier, but takes an instance of 'up.model.state.ROState' instead of the 'assignment' map."""

    def __init__(self, problem: "up.model.problem.Problem"):
        QuantifierSimplifier.__init__(self, problem.env, problem)

    def evaluate(
        self,
        expression: "FNode",
        state: "up.model.state.ROState",
        _variable_assignments: Dict["Expression", "Expression"] = {},
    ):
        """
        Evaluates the given expression in the given state.
        :param expression: The expression that needs to be evaluated.
        :param state: The state where the expression needs to be evaluated.
        :param _variable_assignment: For internal use only. Parameter used to solve quantifiers.
        """
        assert self._problem is not None
        assert self._assignments is None
        assert self._variable_assignments is None
        self._variable_assignments: Optional[
            Dict["Expression", "Expression"]
        ] = _variable_assignments
        self._state = state
        r = self.walk(expression)
        self._variable_assignments = None
        assert r.is_constant()
        return r

    def _deep_subs_simplify(
        self,
        expression: "FNode",
        variables_assignments: Dict["Expression", "Expression"],
    ) -> "FNode":
        """
        This method needs to be updated from the QuantifierRemover in order to use the StateEvaluator inside the
        quantifiers and not the QuantifierSimplifier.
        """
        new_state_evaluator = StateEvaluator(self._problem)
        assert self._variable_assignments is not None
        copy = self._variable_assignments.copy()
        copy.update(variables_assignments)
        r = new_state_evaluator.evaluate(expression, self._state, copy)
        assert r.is_constant()
        return r

    def walk_fluent_exp(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        new_exp = self.manager.FluentExp(expression.fluent(), tuple(args))
        return self._state.get_value(new_exp)

    def walk_param_exp(self, expression: "FNode", args: List["FNode"]) -> "FNode":
        raise UPProblemDefinitionError(
            f"The StateEvaluator.evaluate should only be called on grounded expressions."
        )
